#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
TextRank – is a graph-based ranking model for text processing which can be used in order to find the most relevant sentences in text and also to find keywords.
The algorithm is explained in detail in the paper at https://web.eecs.umich.edu/~mihalcea/papers/mihalcea.emnlp04.pdf

In order to find relevant keywords, the textrank algorithm constructs a word network.
This network is constructed by looking which words follow one another.
A link is set up between two words if they follow one another, the link gets a higher weight if these 2 words occur more frequenctly next to each other in the text.

On top of the resulting network the Pagerank algorithm is applied to get the importance of each word.
The top 1/3 of all these words are kept and are considered relevant.
After this, a keywords table is constructed by combining the relevant words together if they appear following one another in the text.
"""
import re
import typing

from borb.pdf.page.page import Page
from borb.toolkit.text.simple_text_extraction import SimpleTextExtraction
from borb.toolkit.text.stop_words import ENGLISH_STOP_WORDS

try:
    from textblob import TextBlob  # type: ignore [import]
except:
    assert (
        "TextBlob needs to be installed for TextRankKeywordExtraction to work properly."
    )


class TextRankKeywordExtraction(SimpleTextExtraction):
    """
    TextRank – is a graph-based ranking model for text processing which can be used in order to find the most relevant sentences in text and also to find keywords.
    The algorithm is explained in detail in the paper at https://web.eecs.umich.edu/~mihalcea/papers/mihalcea.emnlp04.pdf

    In order to find relevant keywords, the textrank algorithm constructs a word network.
    This network is constructed by looking which words follow one another.
    A link is set up between two words if they follow one another, the link gets a higher weight if these 2 words occur more frequenctly next to each other in the text.

    On top of the resulting network the Pagerank algorithm is applied to get the importance of each word.
    The top 1/3 of all these words are kept and are considered relevant.
    After this, a keywords table is constructed by combining the relevant words together if they appear following one another in the text.
    """

    def __init__(self):
        super().__init__()
        self._stopwords = [x.upper() for x in ENGLISH_STOP_WORDS]

        # keep track of keywords_per_page
        self._keywords_per_page: typing.Dict[
            int, typing.List[typing.Tuple[str, float]]
        ] = {}

    def _end_page(self, page: Page):
        super()._end_page(page)

        # extract text
        txt: str = super(TextRankKeywordExtraction, self).get_text_for_page(
            self._current_page
        )

        # transfer matrix
        mtx: typing.Dict[str, typing.Dict[str, float]] = {}

        # turn txt into lines
        lines = [x for x in re.split("\n*[.?!]+\n*", txt) if len(x) != 0]
        for line in lines:

            # POS tagging
            tags_and_tokens: typing.List[typing.Tuple[str, str]] = TextBlob(line).tags

            # select only NOUN, ADJ
            toks = [x[0] for x in tags_and_tokens if x[1] in ["NN", "JJ"]]

            # build transfer matrix
            for i0 in range(0, len(toks)):
                w0: str = toks[i0].upper()
                if w0 not in mtx:
                    mtx[w0] = {}
                for i1 in range(-3, 3):
                    if i0 + i1 < 0 or i0 + i1 >= len(toks) or i1 == 0:
                        continue
                    w1: str = toks[i0 + i1].upper()
                    mtx[w0][w1] = mtx[w0].get(w1, 0) + 1

        # run eigenvalue algorithm
        ws: typing.List[str] = [x for x, _ in mtx.items()]
        eigenvalues_001: typing.Dict[str, float] = {x: 1 for x in ws}
        eigenvalues_002: typing.Dict[str, float] = {x: 0 for x in ws}
        delta: float = 1
        number_of_iterations: int = 0
        while delta > 0.0001 and number_of_iterations < 128:
            for w0 in ws:
                n: int = len(mtx[w0])
                for w1, f1 in mtx[w0].items():
                    eigenvalues_002[w1] += f1 * (eigenvalues_001[w0] / n)

            # calculate delta
            delta = max([abs(eigenvalues_001[x] - eigenvalues_002[x]) for x in ws])
            number_of_iterations += 1

            # update eigenvalues
            total_weight: float = sum([f for _, f in eigenvalues_002.items()])
            eigenvalues_001 = {
                x: (f / total_weight) for x, f in eigenvalues_002.items()
            }
            eigenvalues_002 = {x: 0 for x, _ in mtx.items()}

        # store keywords
        self._keywords_per_page[self._current_page] = [
            (x, f) for x, f in eigenvalues_001.items()
        ]
        self._keywords_per_page[self._current_page].sort(
            key=lambda x: x[1], reverse=True
        )

    def get_keywords_for_page(self, page_number: int) -> typing.List[typing.Any]:
        """
        This function returns a typing.List[TextRankKeyword] for a given page
        """
        return self._keywords_per_page.get(page_number, [])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A screen annotation (PDF 1.5) specifies a region of a page upon which media clips may be played. It also
serves as an object from which actions can be triggered. 12.6.4.13, “Rendition Actions” discusses the
relationship between screen annotations and rendition actions. Table 187 shows the annotation dictionary
entries specific to this type of annotation.
"""
from borb.pdf.canvas.layout.annotation.annotation import Annotation


class ScreenAnnotation(Annotation):
    """
    A screen annotation (PDF 1.5) specifies a region of a page upon which media clips may be played. It also
    serves as an object from which actions can be triggered. 12.6.4.13, “Rendition Actions” discusses the
    relationship between screen annotations and rendition actions. Table 187 shows the annotation dictionary
    entries specific to this type of annotation.
    """

    def __init__(self):
        super(ScreenAnnotation, self).__init__()
        # TODO



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A redaction annotation (PDF 1.7) identifies content that is intended to be removed from the document. The
intent of redaction annotations is to enable the following process:

a) Content identification. A user applies redact annotations that specify the pieces or regions of content that
should be removed. Up until the next step is performed, the user can see, move and redefine these
annotations.

b) Content removal. The user instructs the viewer application to apply the redact annotations, after which the
content in the area specified by the redact annotations is removed. In the removed content’s place, some
marking appears to indicate the area has been redacted. Also, the redact annotations are removed from
the PDF document.

Redaction annotations provide a mechanism for the first step in the redaction process (content identification).
This allows content to be marked for redaction in a non-destructive way, thus enabling a review process for
evaluating potential redactions prior to removing the specified content.
"""
import typing
import zlib
from decimal import Decimal

from borb.io.read.types import (
    Name,
    List,
    Decimal as bDecimal,
    String,
    Boolean,
    Dictionary,
    Stream,
)
from borb.pdf.canvas.color.color import Color
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.annotation.annotation import Annotation


class RedactAnnotation(Annotation):
    """
    A redaction annotation (PDF 1.7) identifies content that is intended to be removed from the document. The
    intent of redaction annotations is to enable the following process:

    a) Content identification. A user applies redact annotations that specify the pieces or regions of content that
    should be removed. Up until the next step is performed, the user can see, move and redefine these
    annotations.

    b) Content removal. The user instructs the viewer application to apply the redact annotations, after which the
    content in the area specified by the redact annotations is removed. In the removed content’s place, some
    marking appears to indicate the area has been redacted. Also, the redact annotations are removed from
    the PDF document.

    Redaction annotations provide a mechanism for the first step in the redaction process (content identification).
    This allows content to be marked for redaction in a non-destructive way, thus enabling a review process for
    evaluating potential redactions prior to removing the specified content.
    """

    def __init__(
        self,
        bounding_box: Rectangle,
        fill_color: typing.Optional[Color] = None,
        stroke_color: typing.Optional[Color] = None,
        stroke_width: Decimal = Decimal(1),
        overlay_text: typing.Optional[str] = None,
        repeat_overlay_text: typing.Optional[bool] = None,
    ):
        super(RedactAnnotation, self).__init__(bounding_box=bounding_box)

        # (Required) The type of annotation that this dictionary describes; shall
        # be Redact for a redaction annotation.
        self[Name("Subtype")] = Name("Redact")

        # (Optional) An array of three numbers in the range 0.0 to 1.0
        # specifying the components, in the DeviceRGB colour space, of the
        # interior colour with which to fill the redacted region after the affected
        # content has been removed. If this entry is absent, the interior of the
        # redaction region is left transparent. This entry is ignored if the RO
        # entry is present.
        if fill_color is not None:
            self[Name("IC")] = List().set_is_inline(True)  # type: ignore [attr-defined]
            self["IC"].append(bDecimal(fill_color.to_rgb().red))
            self["IC"].append(bDecimal(fill_color.to_rgb().green))
            self["IC"].append(bDecimal(fill_color.to_rgb().blue))

        # (Optional) A text string specifying the overlay text that should be
        # drawn over the redacted region after the affected content has been
        # removed. This entry is ignored if the RO entry is present.
        if overlay_text is not None:
            self[Name("OverlayText")] = String(overlay_text)

        # (Optional) If true, then the text specified by OverlayText should be
        # repeated to fill the redacted region after the affected content has been
        # removed. This entry is ignored if the RO entry is present. Default
        # value: false.
        if repeat_overlay_text is not None:
            assert overlay_text is not None
            self[Name("Repeat")] = Boolean(repeat_overlay_text)

        # (Optional; PDF 1.2) An appearance dictionary specifying how the
        # annotation shall be presented visually on the page (see 12.5.5,
        # “Appearance Streams”). Individual annotation handlers may ignore this
        # entry and provide their own appearances.
        self[Name("AP")] = Dictionary()
        self["AP"][Name("N")] = Stream()
        self["AP"]["N"][Name("Type")] = Name("XObject")
        self["AP"]["N"][Name("Subtype")] = Name("Form")
        appearance_stream_content = "q"
        if stroke_color is not None:
            appearance_stream_content += " %f %f %f RG" % (
                stroke_color.to_rgb().red,
                stroke_color.to_rgb().green,
                stroke_color.to_rgb().blue,
            )
        if fill_color is not None:
            appearance_stream_content += " %f %f %f rg" % (
                fill_color.to_rgb().red,
                fill_color.to_rgb().green,
                fill_color.to_rgb().blue,
            )
        if stroke_color is not None and fill_color is not None:
            appearance_stream_content += " %f w 0 0 100 100 re b" % stroke_width
        elif stroke_color is not None:
            appearance_stream_content += " %f w 0 0 100 100 re s" % stroke_width
        elif fill_color is not None:
            appearance_stream_content += " %f w 0 0 100 100 re f" % stroke_width
        appearance_stream_content += " Q"
        self["AP"]["N"][Name("DecodedBytes")] = bytes(
            appearance_stream_content, "latin1"
        )
        self["AP"]["N"][Name("Bytes")] = zlib.compress(
            self["AP"]["N"][Name("DecodedBytes")]
        )
        self["AP"]["N"][Name("Length")] = bDecimal(len(self["AP"]["N"][Name("Bytes")]))
        self["AP"]["N"][Name("Filter")] = Name("FlateDecode")

        # The lower-left corner of the bounding box (BBox) is set to coordinates (0, 0) in the form coordinate system.
        # The box’s top and right coordinates are taken from the dimensions of the annotation rectangle (the Rect
        # entry in the widget annotation dictionary).
        self["AP"]["N"][Name("BBox")] = List().set_is_inline(True)  # type: ignore [attr-defined]
        self["AP"]["N"]["BBox"].append(bDecimal(0))
        self["AP"]["N"]["BBox"].append(bDecimal(0))
        self["AP"]["N"]["BBox"].append(bDecimal(100))
        self["AP"]["N"]["BBox"].append(bDecimal(100))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of ReadBaseTransformer is responsible for reading Stream objects
"""
import io
import typing
from typing import Any, Optional, Union

from borb.io.filter.stream_decode_util import decode_stream
from borb.io.read.transformer import ReadTransformerState, Transformer
from borb.io.read.types import AnyPDFType, Reference, Stream
from borb.pdf.canvas.event.event_listener import EventListener


class StreamTransformer(Transformer):
    """
    This implementation of ReadBaseTransformer is responsible for reading Stream objects
    """

    def can_be_transformed(
        self, object: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO, AnyPDFType]
    ) -> bool:
        """
        This function returns True if the object to be converted represents a Stream object
        """
        return isinstance(object, Stream)

    def transform(
        self,
        object_to_transform: Union[io.BufferedIOBase, io.RawIOBase, AnyPDFType],
        parent_object: Any,
        context: Optional[ReadTransformerState] = None,
        event_listeners: typing.List[EventListener] = [],
    ) -> Any:
        """
        This function reads a Stream from a byte stream
        """
        assert isinstance(object_to_transform, Stream)
        object_to_transform.set_parent(parent_object)  # type: ignore [attr-defined]

        # resolve references in stream dictionary
        # fmt: off
        assert context is not None, "context must be defined to read Stream objects"
        assert (context.tokenizer is not None), "context.tokenizer must be defined to read Stream objects"
        # fmt: on
        xref = parent_object.get_root().get("XRef")
        for k, v in object_to_transform.items():
            if isinstance(v, Reference):
                v = xref.get_object(v, context.source, context.tokenizer)
                object_to_transform[k] = v

        # apply filter(s)
        object_to_transform = decode_stream(object_to_transform)

        # convert (remainder of) stream dictionary
        for k, v in object_to_transform.items():
            if not isinstance(v, Reference):
                v = self.get_root_transformer().transform(
                    v, object_to_transform, context, []
                )
                if v is not None:
                    object_to_transform[k] = v

        # linkage
        object_to_transform.set_parent(parent_object)  # type: ignore [attr-defined]

        # return
        return object_to_transform



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Set the text rise, T rise , to rise, which shall be a number expressed in
unscaled text space units. Initial value: 0.
"""
import typing
from decimal import Decimal

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class SetTextRise(CanvasOperator):
    """
    Set the text rise, T rise , to rise, which shall be a number expressed in
    unscaled text space units. Initial value: 0.
    """

    def __init__(self):
        super().__init__("Ts", 1)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the Ts operator
        """
        assert isinstance(operands[0], Decimal)
        canvas = canvas_stream_processor.get_canvas()
        canvas.graphics_state.text_rise = operands[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of MultiColumnLayout allows you to specify
a header and footer. In order to render the header and footer
you need to provide their maximum height(s) and a typing.Callable
which will do the actual rendering.
"""
import typing
from decimal import Decimal

from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.page_layout.multi_column_layout import MultiColumnLayout
from borb.pdf.page.page import Page


class HeaderFooterMultiColumnLayout(MultiColumnLayout):
    """
    This implementation of MultiColumnLayout allows you to specify
    a header and footer. In order to render the header and footer
    you need to provide their maximum height(s) and a typing.Callable
    which will do the actual rendering.
    """

    def __init__(
        self,
        page: Page,
        number_of_columns: int = 2,
        horizontal_margin: typing.Optional[Decimal] = None,
        vertical_margin: typing.Optional[Decimal] = None,
        header_callable: typing.Optional[
            typing.Callable[[Page, Rectangle], None]
        ] = None,
        header_height: Decimal = Decimal(12 * 3 * 1.2),
        footer_height: Decimal = Decimal(12 * 3 * 1.2),
        footer_callable: typing.Optional[
            typing.Callable[[Page, Rectangle], None]
        ] = None,
        inter_column_margin: typing.Optional[Decimal] = None,
    ):
        super(HeaderFooterMultiColumnLayout, self).__init__(
            page,
            number_of_columns,
            horizontal_margin,
            vertical_margin,
            inter_column_margin,
        )
        assert header_height >= 0
        assert footer_height >= 0
        self._header_height: Decimal = header_height
        self._footer_height: Decimal = footer_height
        self._header_callable = header_callable
        self._footer_callable = footer_callable

        if self._header_callable is None:
            self._header_height = Decimal(0)
        if self._footer_callable is None:
            self._footer_height = Decimal(0)

        # size of Page is updated to ensure the layout algorithm does not attempt
        # to fill the Page entirely
        assert self._page_height is not None
        self._page_height -= self._header_height
        self._page_height -= self._footer_height

        # modify margins
        self._vertical_margin_bottom += self._footer_height
        self._vertical_margin_top += self._header_height

        # add header and footer to first Page
        self._add_header_footer_to_current_page()

    def _add_header_footer_to_current_page(self) -> None:

        # add header
        if self._header_callable is not None:
            assert self._page_width is not None
            assert self._page_height is not None
            self._header_callable(
                self.get_page(),
                Rectangle(
                    self._horizontal_margin,
                    self._page_height - self._vertical_margin_top,
                    self._page_width - self._horizontal_margin * Decimal(2),
                    self._header_height,
                ),
            )

        # add footer
        if self._footer_callable is not None:
            assert self._page_width is not None
            self._footer_callable(
                self.get_page(),
                Rectangle(
                    self._horizontal_margin,
                    self._vertical_margin_bottom - self._footer_height,
                    self._page_width - self._horizontal_margin * Decimal(2),
                    self._footer_height,
                ),
            )

    def switch_to_next_page(self) -> "PageLayout":  # type: ignore[name-defined]
        """
        This function forces this PageLayout to switch to the nex Page
        This function returns self.
        :return:    self
        """
        # call super
        super(HeaderFooterMultiColumnLayout, self).switch_to_next_page()

        # add header and footer to new page
        self._add_header_footer_to_current_page()

        # return
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    CID-keyed fonts provide a convenient and efficient method for defining multiple-byte character encodings and
   fonts with a large number of glyphs. These capabilities provide great flexibility for representing text in writing
    systems for languages with large character sets, such as Chinese, Japanese, and Korean (CJK).

    The CID-keyed font architecture specifies the external representation of certain font programs, called CMap
    and CIDFont files, along with some conventions for combining and using those files. As mentioned earlier, PDF
    does not support the entire CID-keyed font architecture, which is independent of PDF; CID-keyed fonts may be
    used in other environments.

    A Type 2 CIDFont contains glyph descriptions based on the TrueType font format
"""
import typing

from borb.io.read.types import Decimal as bDecimal
from borb.io.read.types import Name
from borb.pdf.canvas.font.composite_font.cid_font_type_0 import CIDType0Font
from borb.pdf.canvas.font.font import Font


class CIDType2Font(CIDType0Font):
    """
    A Type 2 CIDFont contains glyph descriptions based on the TrueType font format
    """

    def __init__(self):
        super(CIDType2Font, self).__init__()
        self._cid_to_gid_map_cache: typing.Dict[int, int] = {}

    def _empty_copy(self) -> "Font":
        return CIDType2Font()

    def __deepcopy__(self, memodict={}):
        # fmt: off
        f_out: CIDType2Font = super(CIDType2Font, self).__deepcopy__(memodict)
        f_out[Name("Subtype")] = Name("CIDFontType2")
        f_out._width_cache: typing.Dict[int, bDecimal] = {k: v for k, v in self._width_cache.items()}
        return f_out
        # fmt: on



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Set the stroking colour space to DeviceRGB (or the DefaultRGB colour
space; see 8.6.5.6, "Default Colour Spaces") and set the colour to use for
stroking operations. Each operand shall be a number between 0.0
(minimum intensity) and 1.0 (maximum intensity).
"""
import typing
from decimal import Decimal

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.color.color import RGBColor
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class SetRGBStroking(CanvasOperator):
    """
    Set the stroking colour space to DeviceRGB (or the DefaultRGB colour
    space; see 8.6.5.6, "Default Colour Spaces") and set the colour to use for
    stroking operations. Each operand shall be a number between 0.0
    (minimum intensity) and 1.0 (maximum intensity).
    """

    def __init__(self):
        super().__init__("RG", 3)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the RG operator
        """
        # fmt: off
        assert isinstance(operands[0], Decimal), "operand 0 of rg operator must be of type Decimal"
        assert isinstance(operands[1], Decimal), "operand 1 of rg operator must be of type Decimal"
        assert isinstance(operands[2], Decimal), "operand 2 of rg operator must be of type Decimal"
        # fmt: on

        # set stroke_color
        canvas = canvas_stream_processor.get_canvas()
        canvas.graphics_state.stroke_color = RGBColor(
            operands[0], operands[1], operands[2]
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of LayoutElement acts as a common base class to form fields.
"""
import typing

from borb.io.read.types import Dictionary, List, Name
from borb.pdf.canvas.font.font import Font
from borb.pdf.canvas.layout.layout_element import LayoutElement
from borb.pdf.page.page import Page


class FormField(LayoutElement):
    """
    This implementation of LayoutElement acts as a common base class to form fields.
    """

    def _count_fields_on_page(self, page: Page) -> int:
        number_of_fields: int = 0
        acroform_dict: Dictionary = page.get_root()["XRef"]["Trailer"]["Root"].get(  # type: ignore [attr-defined]
            "AcroForm", Dictionary()
        )
        stk: typing.List[typing.Union[Dictionary, List]] = [acroform_dict]
        exp: typing.List[int] = []
        while len(stk) > 0:
            d = stk.pop()
            if id(d) in exp:
                continue
            exp.append(id(d))
            if isinstance(d, Dictionary):
                if "Type" in d and "Subtype" in d and "FT" in d:
                    number_of_fields += 1
                for k, v in d.items():
                    if isinstance(v, Dictionary):
                        stk.append(v)
                    if isinstance(v, List):
                        stk.append(v)
            if isinstance(d, List):
                for c in d:
                    stk.append(c)
        return number_of_fields

    def _get_auto_generated_field_name(self, page: Page) -> str:
        return "field-{0:03d}".format(self._count_fields_on_page(page))

    def _get_font_resource_name(self, font: Font, page: Page):
        # create resources if needed
        if "Resources" not in page:
            page[Name("Resources")] = Dictionary().set_parent(page)  # type: ignore [attr-defined]
        if "Font" not in page["Resources"]:
            page["Resources"][Name("Font")] = Dictionary()

        # insert font into resources
        font_resource_name = [
            k for k, v in page["Resources"]["Font"].items() if v == font
        ]
        if len(font_resource_name) > 0:
            return font_resource_name[0]
        else:
            font_index = len(page["Resources"]["Font"]) + 1
            page["Resources"]["Font"][Name("F%d" % font_index)] = font
            return Name("F%d" % font_index)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of EventListener renders a PDF to a PIL Image
"""
import platform
import typing
from decimal import Decimal
from pathlib import Path

from PIL import Image as PILImage  # type: ignore [import]
from PIL import ImageDraw, ImageFont

from borb.pdf.canvas.color.color import Color
from borb.pdf.page.page_size import PageSize
from borb.pdf.pdf import PDF
from borb.toolkit.export.pdf_to_svg import PDFToSVG


class PDFToJPG(PDFToSVG):
    """
    This implementation of EventListener renders a PDF to a PIL Image
    """

    @staticmethod
    def convert_pdf_to_jpg(file: Path, page_number: int) -> PILImage:  # type: ignore[valid-type]
        """
        This function converts a PDF to an PIL.Image
        """
        l: "PDFToJPG" = PDFToJPG()
        with open(file, "rb") as pdf_file_handle:
            PDF.loads(pdf_file_handle, [l])  # type: ignore [arg-type]
        return l.get_image_for_page(page_number)

    def __init__(
        self,
        default_page_width: Decimal = Decimal(PageSize.A4_PORTRAIT.value[0]),
        default_page_height: Decimal = Decimal(PageSize.A4_PORTRAIT.value[1]),
    ):
        super(PDFToJPG, self).__init__(
            default_page_width=default_page_width,
            default_page_height=default_page_height,
        )
        self._jpg_image_per_page: typing.Dict[int, PILImage] = {}  # type: ignore[valid-type]

        # figure out fonts
        self._regular_font: typing.Optional[Path] = None
        self._bold_font: typing.Optional[Path] = None
        self._italic_font: typing.Optional[Path] = None
        self._bold_italic_font: typing.Optional[Path] = None
        self._find_font_families()

    def _find_font_families(self):
        system: str = platform.system()
        assert system in ["Darwin", "Linux", "Windows"]
        root_font_dir: typing.Optional[Path] = None
        if system == "Linux":
            root_font_dir = Path("/usr/share/fonts")
        if system == "Darwin":
            root_font_dir = Path("/Library/Fonts/")
        if system == "Windows":
            root_font_dir = Path("C:/Windows/Fonts")

        # BFS directory
        ttf_font_files = []
        file_stk: typing.List[Path] = [root_font_dir]
        while len(file_stk) > 0:
            f = file_stk[0]
            file_stk.pop(0)
            if f.is_dir():
                for subdir in f.iterdir():
                    file_stk.append(subdir)
            else:
                if f.name.endswith(".ttf"):
                    ttf_font_files.append(f)

        # find family of fonts
        for c in ["LiberationSans", "LiberationMono"]:
            suffixes = ["-Regular", "-Italic", "-Bold", "-BoldItalic"]
            all_fonts_present = all(
                [
                    y in [x.name for x in ttf_font_files]
                    for y in [c + x + ".ttf" for x in suffixes]
                ]
            )
            if all_fonts_present:
                self._regular_font = [
                    x for x in ttf_font_files if x.name.endswith(c + "-Regular.ttf")
                ][0]
                self._bold_font = [
                    x for x in ttf_font_files if x.name.endswith(c + "-Bold.ttf")
                ][0]
                self._italic_font = [
                    x for x in ttf_font_files if x.name.endswith(c + "-Italic.ttf")
                ][0]
                self._bold_italic_font = [
                    x for x in ttf_font_files if x.name.endswith(c + "-BoldItalic.ttf")
                ][0]

    def _begin_page(
        self, page_nr: Decimal, page_width: Decimal, page_height: Decimal
    ) -> None:
        self._jpg_image_per_page[int(page_nr)] = PILImage.new(
            "RGB", (int(page_width), int(page_height)), color=(255, 255, 255)
        )

    def _render_text(
        self,
        page_nr: Decimal,
        page_width: Decimal,
        page_height: Decimal,
        x: Decimal,
        y: Decimal,
        font_color: Color,
        font_size: Decimal,
        font_name: str,
        bold: bool,
        italic: bool,
        text: str,
    ):
        if len(text.strip()) == 0:
            return

        assert self._bold_font
        assert self._bold_italic_font
        assert self._italic_font
        assert self._regular_font

        font_path = self._regular_font
        if bold and italic:
            font_path = self._bold_italic_font
        elif bold:
            font_path = self._bold_font
        elif italic:
            font_path = self._italic_font

        # instantiate font
        font = ImageFont.truetype(str(font_path), int(font_size))

        # draw text
        assert self._jpg_image_per_page.get(int(page_nr)) is not None
        draw = ImageDraw.Draw(self._jpg_image_per_page[int(page_nr)])
        draw.text(
            (float(x), float(page_height - y)),
            text,
            font=font,
            fill=(
                int(font_color.to_rgb().red),
                int(font_color.to_rgb().green),
                int(font_color.to_rgb().blue),
            ),
        )

    def _render_image(
        self,
        page_nr: Decimal,
        page_width: Decimal,
        page_height: Decimal,
        x: Decimal,
        y: Decimal,
        image_width: Decimal,
        image_height: Decimal,
        image: PILImage,  # type: ignore[valid-type]
    ):
        page_image = self._jpg_image_per_page.get(int(page_nr))
        assert page_image is not None

        # resize
        image = image.resize((int(image_width), int(image_height)))

        # paste
        page_image.paste(image, (int(x), int(page_height - y - image_height)))

    def get_image_for_page(self, page_nr: int) -> PILImage:  # type: ignore[valid-type]
        """
        This function returns the PIL.Image for a given page_nr
        """
        assert page_nr in self._jpg_image_per_page
        return self._jpg_image_per_page[page_nr]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Begin a marked-content sequence terminated by a balancing EMC
operator. tag shall be a name object indicating the role or significance of
the sequence.
"""

import typing

from borb.io.read.types import AnyPDFType, Name
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class BeginMarkedContent(CanvasOperator):
    """
    Begin a marked-content sequence terminated by a balancing EMC
    operator. tag shall be a name object indicating the role or significance of
    the sequence.
    """

    def __init__(self):
        super().__init__("BMC", 1)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the BMC operator
        """
        assert isinstance(operands[0], Name), "Operand 0 of BMC must be a Name"
        canvas = canvas_stream_processor.get_canvas()
        canvas.marked_content_stack.append(operands[0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of EventListener renders a PDF to an SVG image
"""
import base64
import io
import typing
import xml.etree.ElementTree as ET
from decimal import Decimal
from pathlib import Path

from PIL import Image as PILImage  # type: ignore [import]

from borb.pdf.canvas.color.color import Color
from borb.pdf.canvas.event.begin_page_event import BeginPageEvent
from borb.pdf.canvas.event.chunk_of_text_render_event import ChunkOfTextRenderEvent
from borb.pdf.canvas.event.event_listener import Event, EventListener
from borb.pdf.canvas.event.image_render_event import ImageRenderEvent
from borb.pdf.page.page import Page
from borb.pdf.page.page_size import PageSize
from borb.pdf.pdf import PDF


class PDFToSVG(EventListener):
    """
    This implementation of EventListener renders a PDF to an SVG image
    """

    @staticmethod
    def convert_pdf_to_svg(file: Path, page_number: int) -> ET.Element:
        """
        This function converts a PDF to an SVG ET.Element
        """
        l: "PDFToSVG" = PDFToSVG()
        with open(file, "rb") as pdf_file_handle:
            PDF.loads(pdf_file_handle, [l])  # type: ignore [arg-type]
        return l.get_image_for_page(page_number)

    def __init__(
        self,
        default_page_width: Decimal = Decimal(PageSize.A4_PORTRAIT.value[0]),
        default_page_height: Decimal = Decimal(PageSize.A4_PORTRAIT.value[1]),
    ):
        self._default_page_width = default_page_width
        self._default_page_height = default_page_height
        self._page: typing.Optional[Page] = None
        self._page_nr = Decimal(-1)
        self._svg_per_page: typing.Dict[int, ET.Element] = {}

    def _event_occurred(self, event: Event) -> None:
        # BeginPageEvent
        if isinstance(event, BeginPageEvent):
            self._page_nr += Decimal(1)
            self._page = event.get_page()
            self._begin_page(
                self._page_nr,
                self._page.get_page_info().get_width() or self._default_page_width,
                self._page.get_page_info().get_height() or self._default_page_height,
            )
        # ImageRenderEvent
        if isinstance(event, ImageRenderEvent):
            assert self._page is not None
            self._render_image(
                self._page_nr,
                self._page.get_page_info().get_width() or self._default_page_width,
                self._page.get_page_info().get_height() or self._default_page_height,
                event.get_x(),
                event.get_y(),
                event.get_width(),
                event.get_height(),
                event.get_image(),
            )
        # ChunkOfTextRenderEvent
        if isinstance(event, ChunkOfTextRenderEvent):
            assert self._page is not None
            font_name_as_str = "Helvetica"
            if event._font.get_font_name():
                font_name_as_str = str(event._font.get_font_name())
            self._render_text(
                self._page_nr,
                self._page.get_page_info().get_width() or self._default_page_width,
                self._page.get_page_info().get_height() or self._default_page_height,
                event.get_baseline().get_x(),
                event.get_baseline().get_y(),
                event._font_color,
                event._font_size,
                font_name_as_str.replace("#20", " ")
                .replace(",Bold", "")
                .replace(",bold", "")
                .replace("Bold", "")
                .replace("bold", "")
                .replace(",Italic", "")
                .replace(",italic", "")
                .replace("Italic", "")
                .replace("italic", ""),
                "BOLD" in font_name_as_str.upper(),
                "ITALIC" in font_name_as_str.upper(),
                event._text,
            )

    def _begin_page(
        self, page_nr: Decimal, page_width: Decimal, page_height: Decimal
    ) -> None:

        # init svg image
        ET.register_namespace("", "http://www.w3.org/2000/svg")
        svg_element = ET.Element("svg")
        svg_element.set("xmlns:xlink", "http://www.w3.org/1999/xlink")
        svg_element.set("viewbox", "0 0 %d %d" % (page_width, page_height))

        # white background
        rct_element = ET.Element("rect")
        rct_element.set("width", str(page_width))
        rct_element.set("height", str(page_height))
        rct_element.set("style", "fill:rgb(255, 255, 255);")
        svg_element.append(rct_element)
        self._svg_per_page[int(page_nr)] = svg_element  # type: ignore [assignment]

    def _render_text(
        self,
        page_nr: Decimal,
        page_width: Decimal,
        page_height: Decimal,
        x: Decimal,
        y: Decimal,
        font_color: Color,
        font_size: Decimal,
        font_name: str,
        bold: bool,
        italic: bool,
        text: str,
    ):
        if len(text.strip()) == 0:
            return

        text_element = ET.Element("text")

        # bold
        if bold:
            text_element.set("_font-weight", "bold")

        # italic
        if italic:
            text_element.set("_font-style", "italic")

        # font_color, font_size, preserve space
        font_color_rgb = font_color.to_rgb()
        text_element.set(
            "style",
            "fill:rgb(%d, %d, %d); _font-size:%d px; white-space: pre;"
            % (
                font_color_rgb.red,
                font_color_rgb.green,
                font_color_rgb.blue,
                int(font_size),
            ),
        )
        text_element.set("xml:space", "preserve")

        # set font-family
        text_element.set("_font-family", font_name)

        # text
        text_element.text = text

        # position
        text_element.set("x", str(int(x)))
        text_element.set("y", str(int(page_height - y)))

        # append
        assert self._svg_per_page[int(page_nr)] is not None
        self._svg_per_page[int(page_nr)].append(text_element)

    def _render_image(
        self,
        page_nr: Decimal,
        page_width: Decimal,
        page_height: Decimal,
        x: Decimal,
        y: Decimal,
        image_width: Decimal,
        image_height: Decimal,
        image: PILImage,  # type: ignore[valid-type]
    ):
        pass

        # base64 image
        with io.BytesIO() as output:
            image.convert("RGB").save(output, format="JPEG")  # type: ignore[attr-defined]
            base64_image = "data:image/png;base64," + base64.b64encode(
                output.getvalue()
            ).decode("utf-8")

        image_element = ET.Element("image")
        image_element.set("width", str(int(image_width)))
        image_element.set("height", str(int(image_height)))
        image_element.set("xlink:href", base64_image)

        # position
        image_element.set("x", str(int(x)))
        image_element.set("y", str(int(page_height - y - image_height)))

        # append
        assert self._svg_per_page[int(page_nr)] is not None
        self._svg_per_page[int(page_nr)].append(image_element)

    def get_image_for_page(self, page_nr: int) -> ET.Element:
        """
        This function returns the ET.Element for a given page_nr
        """
        assert page_nr in self._svg_per_page
        return self._svg_per_page[page_nr]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
(PDF 1.1) Same as CS but used for nonstroking operations.
"""
import typing
from decimal import Decimal
from typing import List

from borb.io.read.types import AnyPDFType, Name
from borb.pdf.canvas.color.color import CMYKColor, GrayColor, RGBColor, Separation
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class SetColorSpaceNonStroking(CanvasOperator):
    """
    (PDF 1.1) Same as CS but used for nonstroking operations.
    """

    def __init__(self):
        super().__init__("cs", 1)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the cs operator
        """
        assert isinstance(operands[0], Name), "Operand 0 of cs must be a Name"
        color_space_name: Name = operands[0]
        color_space: List = []

        if color_space_name not in [
            "DeviceGray",
            "DeviceRGB",
            "DeviceCMYK",
            "CalGray",
            "CalRGB",
            "Lab",
            "ICCBased",
            "Indexed",
            "Pattern",
            "Separation",
        ]:
            color_space_name = canvas_stream_processor.get_resource(
                "ColorSpace", color_space_name
            )

        if not isinstance(color_space_name, Name) and isinstance(
            color_space_name, List
        ):
            assert isinstance(color_space_name[0], Name)
            color_space = color_space_name
            color_space_name = color_space_name[0]

        #
        # Device
        #
        canvas = canvas_stream_processor.get_canvas()
        if color_space_name == "DeviceGray":
            canvas.graphics_state.non_stroke_color_space = color_space_name
            canvas.graphics_state.non_stroke_color = GrayColor(Decimal(0))
            return

        if color_space_name == "DeviceRGB":
            canvas.graphics_state.non_stroke_color_space = color_space_name
            canvas.graphics_state.non_stroke_color = RGBColor(
                Decimal(0), Decimal(0), Decimal(0)
            )
            return

        if color_space_name == "DeviceCMYK":
            canvas.graphics_state.non_stroke_color_space = color_space_name
            canvas.graphics_state.non_stroke_color = CMYKColor(
                Decimal(0), Decimal(0), Decimal(0), Decimal(1)
            )
            return

        #
        # CIE-based
        #

        if color_space_name == "CalGray":
            # fmt: off
            canvas.graphics_state.non_stroke_color_space = color_space_name
            canvas.graphics_state.non_stroke_color = GrayColor(Decimal(0))
            return
            # fmt: on
        if color_space_name == "CalRGB":
            # fmt: off
            canvas.graphics_state.non_stroke_color_space = color_space_name
            canvas.graphics_state.non_stroke_color = RGBColor(Decimal(0), Decimal(0), Decimal(0))
            return
            # fmt: on
        if color_space_name == "Lab":
            canvas.graphics_state.non_stroke_color_space = color_space_name
            return
        if color_space_name == "ICCBased":
            # fmt: off
            canvas.graphics_state.non_stroke_color_space = color_space_name
            canvas.graphics_state.non_stroke_color = RGBColor(Decimal(0), Decimal(0), Decimal(0))
            return
            # fmt: on

        #
        # Special
        #

        if color_space_name == "Indexed":
            canvas.graphics_state.non_stroke_color_space = color_space_name
            return
        if operands[0] == "Pattern":
            canvas.graphics_state.non_stroke_color_space = operands[0]
            return
        if color_space_name == "Separation":
            # fmt: off
            canvas.graphics_state.non_stroke_color_space = color_space
            canvas.graphics_state.non_stroke_color = Separation(color_space, [Decimal(0)])
            # fmt: on



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This class processes a given bytes stream of instructions on a Canvas and Page.
By abstracting this logic into one class, we can easily change the resources it uses
(which is needed to handle /Form XObjects).
"""
import io
import logging
import os
import time
import typing

from borb.io.read.tokenize.high_level_tokenizer import HighLevelTokenizer
from borb.io.read.types import AnyPDFType, CanvasOperatorName, Dictionary
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator
from borb.pdf.canvas.operator.color.set_cmyk_non_stroking import SetCMYKNonStroking
from borb.pdf.canvas.operator.color.set_cmyk_stroking import SetCMYKStroking
from borb.pdf.canvas.operator.color.set_color_non_stroking import SetColorNonStroking
from borb.pdf.canvas.operator.color.set_color_space_non_stroking import (
    SetColorSpaceNonStroking,
)
from borb.pdf.canvas.operator.color.set_color_space_stroking import (
    SetColorSpaceStroking,
)
from borb.pdf.canvas.operator.color.set_color_stroking import SetColorStroking
from borb.pdf.canvas.operator.color.set_gray_non_stroking import SetGrayNonStroking
from borb.pdf.canvas.operator.color.set_gray_stroking import SetGrayStroking
from borb.pdf.canvas.operator.color.set_rgb_non_stroking import SetRGBNonStroking
from borb.pdf.canvas.operator.color.set_rgb_stroking import SetRGBStroking
from borb.pdf.canvas.operator.compatibility.begin_compatibility_section import (
    BeginCompatibilitySection,
)
from borb.pdf.canvas.operator.compatibility.end_compatibility_section import (
    EndCompatibilitySection,
)
from borb.pdf.canvas.operator.marked_content.begin_marked_content import (
    BeginMarkedContent,
)
from borb.pdf.canvas.operator.marked_content.begin_marked_content_with_property_list import (
    BeginMarkedContentWithPropertyList,
)
from borb.pdf.canvas.operator.marked_content.end_marked_content import EndMarkedContent
from borb.pdf.canvas.operator.path_construction.append_cubic_bezier import (
    AppendCubicBezierCurve1,
    AppendCubicBezierCurve2,
    AppendCubicBezierCurve3,
)
from borb.pdf.canvas.operator.path_construction.append_line_segment import (
    AppendLineSegment,
)
from borb.pdf.canvas.operator.path_construction.begin_subpath import BeginSubpath
from borb.pdf.canvas.operator.path_construction.close_subpath import CloseSubpath
from borb.pdf.canvas.operator.path_painting.close_and_stroke_path import (
    CloseAndStrokePath,
)
from borb.pdf.canvas.operator.path_painting.stroke_path import StrokePath
from borb.pdf.canvas.operator.state.modify_transformation_matrix import (
    ModifyTransformationMatrix,
)
from borb.pdf.canvas.operator.state.pop_graphics_state import PopGraphicsState
from borb.pdf.canvas.operator.state.push_graphics_state import PushGraphicsState
from borb.pdf.canvas.operator.state.set_line_width import SetLineWidth
from borb.pdf.canvas.operator.text.begin_text import BeginTextObject
from borb.pdf.canvas.operator.text.end_text import EndTextObject
from borb.pdf.canvas.operator.text.move_text_position import MoveTextPosition
from borb.pdf.canvas.operator.text.move_text_position_set_leading import (
    MoveTextPositionSetLeading,
)
from borb.pdf.canvas.operator.text.move_to_next_line import MoveToNextLine
from borb.pdf.canvas.operator.text.move_to_next_line_show_text import (
    MoveToNextLineShowText,
)
from borb.pdf.canvas.operator.text.set_character_spacing import SetCharacterSpacing
from borb.pdf.canvas.operator.text.set_font_and_size import SetFontAndSize
from borb.pdf.canvas.operator.text.set_horizontal_text_scaling import (
    SetHorizontalScaling,
)
from borb.pdf.canvas.operator.text.set_spacing_move_to_next_line_show_text import (
    SetSpacingMoveToNextLineShowText,
)
from borb.pdf.canvas.operator.text.set_text_leading import SetTextLeading
from borb.pdf.canvas.operator.text.set_text_matrix import SetTextMatrix
from borb.pdf.canvas.operator.text.set_text_rendering_mode import SetTextRenderingMode
from borb.pdf.canvas.operator.text.set_text_rise import SetTextRise
from borb.pdf.canvas.operator.text.set_word_spacing import SetWordSpacing
from borb.pdf.canvas.operator.text.show_text import ShowText
from borb.pdf.canvas.operator.text.show_text_with_glyph_positioning import (
    ShowTextWithGlyphPositioning,
)
from borb.pdf.canvas.operator.xobject.do import Do

logger = logging.getLogger(__name__)


class CanvasStreamProcessor:
    """
    This class processes a given bytes stream of instructions on a Canvas and Page.
    By abstracting this logic into one class, we can easily change the resources it uses
    (which is needed to handle /Form XObjects).
    """

    def __init__(
        self,
        page: "Page",  # type: ignore[name-defined]
        canvas: "Canvas",  # type: ignore[name-defined]
        resource_dictionaries: typing.List[Dictionary] = [],
    ):
        self._page: "Page" = page  # type: ignore[name-defined]
        self._canvas: "Canvas" = canvas  # type: ignore[name-defined]
        self._resource_dictionaries: typing.List[Dictionary] = resource_dictionaries

        # initialize operators
        self._canvas_operators: typing.Dict[str, CanvasOperator] = {
            x.get_text(): x
            for x in [
                # color
                SetCMYKNonStroking(),
                SetCMYKStroking(),
                SetColorNonStroking(self),
                SetColorStroking(self),
                SetGrayNonStroking(),
                SetGrayStroking(),
                SetRGBNonStroking(),
                SetRGBStroking(),
                SetColorSpaceStroking(),
                SetColorSpaceNonStroking(),
                # compatibility
                BeginCompatibilitySection(),
                EndCompatibilitySection(),
                # marked content
                BeginMarkedContent(),
                BeginMarkedContentWithPropertyList(),
                EndMarkedContent(),
                # path construction
                AppendCubicBezierCurve1(),
                AppendCubicBezierCurve2(),
                AppendCubicBezierCurve3(),
                AppendLineSegment(),
                BeginSubpath(),
                CloseSubpath(),
                # path painting
                CloseAndStrokePath(),
                StrokePath(),
                # state
                ModifyTransformationMatrix(),
                PopGraphicsState(),
                PushGraphicsState(),
                SetLineWidth(),
                # text
                BeginTextObject(),
                EndTextObject(),
                MoveTextPosition(),
                MoveTextPositionSetLeading(),
                MoveToNextLineShowText(),
                MoveToNextLine(),
                SetCharacterSpacing(),
                SetFontAndSize(),
                SetHorizontalScaling(),
                SetSpacingMoveToNextLineShowText(),
                SetTextLeading(),
                SetTextMatrix(),
                SetTextRenderingMode(),
                SetTextRise(),
                SetWordSpacing(),
                ShowText(),
                ShowTextWithGlyphPositioning(),
                # xobject
                Do(),
            ]
        }

    def create_child_canvas_stream_processor(
        self, resource_dictionaries: typing.List[Dictionary]
    ) -> "CanvasStreamProcessor":
        """
        This function creates a (child) CanvasStreamProcessor.
        The child processor will have the same resource dictionaries (fonts, images, etc) as its parent (self),
        but can optionally add more resources (such as when a content stream is defined in an XObject).
        """
        return CanvasStreamProcessor(
            self._page,
            self._canvas,
            self._resource_dictionaries + resource_dictionaries,
        )

    def get_operator(self, name: str) -> typing.Optional["CanvasOperator"]:  # type: ignore [name-defined]
        """
        This function returns the CanvasOperator matching the given operator-name.
        This allows operator re-use between different implementations of Canvas
        """
        return self._canvas_operators.get(name)

    def get_page(self) -> "Page":  # type: ignore[name-defined]
        """
        This function returns the Page on which this CanvasStreamProcessor is active.
        """
        return self._page

    def get_canvas(self) -> "Canvas":  # type: ignore[name-defined]
        """
        This function returns the Canvas on which this CanvasStreamProcessor is active.
        """
        return self._canvas

    def get_resource(
        self, resource_type_name: str, name: str
    ) -> typing.Optional[typing.Any]:
        """
        This functions looks up a resource (e.g. Font, Image, XObject) in the given resource hierarchy.
        e.g. for a content stream in an XObject, first its own /Resources entry is tried, and lastly the Page resources.
        """
        # check external Resources
        for i in range(len(self._resource_dictionaries) - 1, -1, -1):
            rd = self._resource_dictionaries[i]
            if (resource_type_name in rd) and (name in rd[resource_type_name]):
                return rd[resource_type_name][name]
        # check Page[Resources]
        if (
            resource_type_name in self._page["Resources"]
            and name in self._page["Resources"][resource_type_name]
        ):
            return self._page["Resources"][resource_type_name][name]
        return None

    def read(
        self,
        io_source: typing.Union[io.BytesIO, io.IOBase],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> "CanvasStreamProcessor":  # type: ignore [name-defined]
        """
        This method reads a byte stream of canvas operators, and processes them, returning this Canvas afterwards
        """
        io_source.seek(0, os.SEEK_END)
        length = io_source.tell()
        io_source.seek(0)

        canvas_tokenizer = HighLevelTokenizer(io_source)

        # process content
        operand_stk: typing.List[AnyPDFType] = []
        instruction_number: int = 0
        time_per_operator: typing.Dict[str, float] = {}
        calls_per_operator: typing.Dict[str, int] = {}
        while canvas_tokenizer.tell() != length:

            # print("<canvas pos='%d' length='%d' percentage='%d'/>" % ( canvas_tokenizer.tell(), length, int(canvas_tokenizer.tell() * 100 / length)))

            # attempt to read object
            tell_before: int = canvas_tokenizer.tell()
            obj = canvas_tokenizer.read_object()
            tell_after: int = canvas_tokenizer.tell()
            if obj is None and tell_before == tell_after:
                break

            # push argument onto stack
            if not isinstance(obj, CanvasOperatorName):
                operand_stk.append(obj)
                continue

            # process operator
            instruction_number += 1
            operator = self._canvas_operators.get(str(obj), None)
            if operator is None:
                logger.debug("Missing operator %s" % obj)
                continue

            if not self._canvas.in_compatibility_section:
                assert len(operand_stk) >= operator.get_number_of_operands()
            operands: typing.List[AnyPDFType] = []  # type: ignore [name-defined]
            for _ in range(0, operator.get_number_of_operands()):
                operands.insert(0, operand_stk.pop(-1))

            # invoke
            try:
                on: str = operator.get_text()
                if on not in time_per_operator:
                    time_per_operator[on] = 0
                if on not in calls_per_operator:
                    calls_per_operator[on] = 1
                else:
                    calls_per_operator[on] += 1
                delta: float = time.time()
                operator.invoke(self, operands, event_listeners)
                delta = time.time() - delta
                time_per_operator[on] += delta

            except Exception as e:
                if not self._canvas.in_compatibility_section:
                    raise e

        # return
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This function decodes a Stream, applying the filters specified in the Filter entry
of its stream dictionary
"""
import typing

from borb.io.filter.ascii85_decode import ASCII85Decode
from borb.io.filter.flate_decode import FlateDecode
from borb.io.filter.lzw_decode import LZWDecode
from borb.io.filter.run_length_decode import RunLengthDecode
from borb.io.read.types import Decimal, Dictionary, List, Name, Stream


def decode_stream(s: Stream) -> Stream:
    """
    This function decodes a Stream, applying the filters specified in the Filter entry
    of its stream dictionary
    """
    # fmt: off
    assert isinstance(s, Stream), "decode_stream only works on Stream objects"
    assert ("Bytes" in s), "decode_stream only works on Stream objects with a `Bytes` key."
    # fmt: on

    # determine filter(s) to apply
    filters: typing.List[str] = []
    if "Filter" in s:
        if isinstance(s["Filter"], List):
            filters = s["Filter"]
        else:
            filters = [s["Filter"]]

    decode_params: typing.List[Dictionary] = []
    if "DecodeParms" in s:
        if isinstance(s["DecodeParms"], List):
            decode_params = s["DecodeParms"]
            decode_params = [x or Dictionary() for x in decode_params]
        else:
            assert s["DecodeParms"] is not None
            assert isinstance(s["DecodeParms"], Dictionary)
            decode_params = [s["DecodeParms"]]
    else:
        decode_params = [Dictionary() for x in range(0, len(filters))]

    # apply filter(s)
    transformed_bytes = s["Bytes"]
    for filter_index, filter_name in enumerate(filters):
        # FLATE
        if filter_name in ["FlateDecode", "Fl"]:
            transformed_bytes = FlateDecode.decode(
                bytes_in=transformed_bytes,
                columns=int(decode_params[filter_index].get("Columns", Decimal(1))),
                predictor=int(decode_params[filter_index].get("Predictor", Decimal(1))),
                bits_per_component=int(
                    decode_params[filter_index].get("BitsPerComponent", Decimal(8))
                ),
            )
            continue

        # ASCII85
        if filter_name in ["ASCII85Decode"]:
            transformed_bytes = ASCII85Decode.decode(transformed_bytes)
            continue

        # LZW
        if filter_name in ["LZWDecode"]:
            transformed_bytes = LZWDecode().decode(transformed_bytes)
            continue

        # RunLengthDecode
        if filter_name in ["RunLengthDecode"]:
            transformed_bytes = RunLengthDecode.decode(transformed_bytes)
            continue

        # unknown filter
        assert False, "Unknown /Filter %s" % filter_name

    # set DecodedBytes
    s[Name("DecodedBytes")] = transformed_bytes

    # set Type if not yet set
    # if "Type" not in s:
    #    s[Name("Type")] = Name("Stream")

    # return
    return s



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Same as G but used for nonstroking operations.
"""
import typing
from decimal import Decimal

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.color.color import GrayColor
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class SetGrayNonStroking(CanvasOperator):
    """
    Same as G but used for nonstroking operations.
    """

    def __init__(self):
        super().__init__("g", 1)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the g operator
        """
        assert isinstance(operands[0], Decimal), "Operand 0 of g must be a Decimal"
        canvas = canvas_stream_processor.get_canvas()
        canvas.graphics_state.non_stroke_color = GrayColor(operands[0])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of ReadBaseTransformer is responsible for reading the XRef object
"""
import io
import os
import typing
from decimal import Decimal
from typing import Any, Optional, Union

from borb.io.read.encryption.standard_security_handler import StandardSecurityHandler
from borb.io.read.tokenize.high_level_tokenizer import HighLevelTokenizer
from borb.io.read.transformer import ReadTransformerState, Transformer
from borb.io.read.types import AnyPDFType, Dictionary, Name
from borb.pdf.canvas.event.event_listener import EventListener, Event
from borb.pdf.document.document import Document
from borb.pdf.xref.plaintext_xref import PlainTextXREF
from borb.pdf.xref.rebuilt_xref import RebuiltXREF
from borb.pdf.xref.stream_xref import StreamXREF
from borb.pdf.xref.xref import XREF


class BeginDocumentEvent(Event):
    """
    This implementation of Event is triggered right before the Document is processed
    """

    pass


class EndDocumentEvent(Event):
    """
    This implementation of Event is triggered right after the Document was processed
    """

    pass


class XREFTransformer(Transformer):
    """
    This implementation of ReadBaseTransformer is responsible for reading the XRef object
    """

    def can_be_transformed(
        self, object: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO, AnyPDFType]
    ) -> bool:
        """
        This function returns True if the object to be converted represents a cross-reference table
        """
        return isinstance(object, io.IOBase)

    def transform(
        self,
        object_to_transform: Union[io.BufferedIOBase, io.RawIOBase, AnyPDFType],
        parent_object: Any,
        context: Optional[ReadTransformerState] = None,
        event_listeners: typing.List[EventListener] = [],
    ) -> Any:
        """
        This function reads a cross-reference table from a byte stream
        """

        # update context
        assert context is not None
        assert isinstance(object_to_transform, io.BufferedIOBase) or isinstance(
            object_to_transform, io.RawIOBase
        )
        context.root_object = Document()
        context.source = object_to_transform
        context.tokenizer = HighLevelTokenizer(context.source)

        # add listener(s)
        for l in event_listeners:
            l._event_occurred(BeginDocumentEvent())  # type: ignore [attr-defined]

        # remove prefix
        XREFTransformer._remove_prefix(context)

        # check header
        XREFTransformer._check_header(context)

        # file size
        context.source.seek(0, os.SEEK_END)
        file_length = context.source.tell()
        context.source.seek(0)
        context.root_object[Name("FileSize")] = Decimal(file_length)

        # build XREF object
        self._read_xref(context)

        # transform trailer dictionary
        xref = context.root_object.get("XRef")
        assert xref is not None
        assert isinstance(xref, XREF)

        # check for password protected PDF
        if "Trailer" in xref and "Encrypt" in xref["Trailer"]:

            # transform /Encrypt dictionary
            # fmt: off
            xref["Trailer"][Name("Encrypt")] = self.get_root_transformer().transform(xref["Trailer"]["Encrypt"], xref["Trailer"], context, event_listeners)
            # fmt: on

            # build encryption handler
            v: int = int(xref["Trailer"]["Encrypt"].get("V", Decimal(0)))
            r: int = int(xref["Trailer"]["Encrypt"]["R"])
            if r != 2 and r != 3:
                assert (
                    False
                ), "R is not 2 or 3. A number specifying which revision of the standard security handler shall be used to interpret this dictionary."
            if v == 0:
                assert False, (
                    "V is 0. An algorithm that is undocumented. "
                    "This value shall not be used."
                )
            if v == 1:
                context.security_handler = StandardSecurityHandler(
                    xref["Trailer"]["Encrypt"], context.password
                )
            if v == 2:
                context.security_handler = StandardSecurityHandler(
                    xref["Trailer"]["Encrypt"], context.password
                )
            if v == 3:
                assert False, (
                    "V is 3. (PDF 1.4) An unpublished algorithm that permits encryption key lengths ranging from 40 to 128 bits. "
                    "This value shall not appear in a conforming PDF file."
                )
            if v == 4:
                assert False, "V is 4. Currently unsupported encryption dictionary."

            # raise error
            raise NotImplementedError(
                "password-protected PDFs are currently not supported"
            )

        # transform /Trailer
        trailer = self.get_root_transformer().transform(
            context.root_object["XRef"]["Trailer"],
            context.root_object,
            context,
            event_listeners,
        )

        assert trailer is not None
        assert isinstance(trailer, Dictionary)
        xref[Name("Trailer")] = trailer
        for k in ["DecodeParms", "Filter", "Index", "Length", "Prev", "W"]:
            if k in xref["Trailer"]:
                xref["Trailer"].pop(k)

        # notify
        for l in event_listeners:
            l._event_occurred(EndDocumentEvent())  # type: ignore [attr-defined]

        # return
        return context.root_object

    @staticmethod
    def _remove_prefix(context: ReadTransformerState) -> None:

        assert context is not None
        assert context.source is not None
        assert context.tokenizer is not None

        # read first 2 Kb
        bytes_near_sof = context.source.read(2048)
        assert bytes_near_sof is not None
        text_near_sof = bytes_near_sof.decode("latin-1")
        context.source.seek(0)

        # find %PDF-
        index_of_pdf_comment = -1
        try:
            index_of_pdf_comment = text_near_sof.index("%PDF")
        except ValueError:
            pass

        # truncate
        if index_of_pdf_comment > 0:
            # determine end of file
            end_of_file = context.source.seek(0, os.SEEK_END)
            context.source.seek(0)
            # build byte array
            bts = context.source.read(end_of_file)
            assert bts is not None
            bts = bts[index_of_pdf_comment:end_of_file]
            # reset values in context
            context.source = io.BytesIO(bts)
            context.tokenizer._io_source = context.source

    @staticmethod
    def _check_header(context: ReadTransformerState) -> None:
        """
        This function checks whether or not the first bytes in the document contain the text %PDF
        :param context: the TransformerContext (containing the io source)
        :type context: TransformerContext
        """
        assert context is not None
        assert context.source is not None
        assert context.tokenizer is not None
        context.source.seek(0)
        arr = [
            t
            for t in [context.tokenizer.next_token() for _ in range(0, 10)]
            if t is not None
        ]
        assert len(arr) > 0
        assert any([t.get_text().startswith("%PDF") for t in arr])

    def _read_xref(
        self, context: ReadTransformerState, initial_offset: Optional[int] = None
    ) -> None:
        """
        This function attempts to read the XREF table, first as plaintext, then as a stream
        :param context:         the TransformerContext (containing the io source)
        :type context:          TransformerContext
        :param initial_offset:  the initial byte offset at which to read (set to None to allow this method to find the XREF)
        :type initial_offset:   int
        """
        assert context is not None
        assert context.root_object is not None
        assert context.source is not None
        assert context.tokenizer is not None

        doc = context.root_object
        src = context.source
        tok = context.tokenizer

        most_recent_xref: Optional[XREF] = None
        exceptions_to_rethrow = []

        # attempt to read plaintext XREF
        try:
            most_recent_xref = PlainTextXREF()
            assert most_recent_xref is not None
            most_recent_xref.set_parent(doc)  # type: ignore [attr-defined]
            most_recent_xref.read(src, tok, initial_offset)
            if "XRef" in doc:
                doc[Name("XRef")] = doc["XRef"].merge(most_recent_xref)
            else:
                doc[Name("XRef")] = most_recent_xref
        except Exception as ex0:
            most_recent_xref = None
            exceptions_to_rethrow.append(ex0)

        # attempt to read stream XREF
        if most_recent_xref is None:
            try:
                most_recent_xref = StreamXREF()
                assert most_recent_xref is not None
                most_recent_xref.set_parent(doc)  # type: ignore [attr-defined]
                most_recent_xref.read(src, tok, initial_offset)
                if "XRef" in doc:
                    doc[Name("XRef")] = doc["XRef"].merge(most_recent_xref)
                else:
                    doc[Name("XRef")] = most_recent_xref
            except Exception as ex0:
                most_recent_xref = None
                exceptions_to_rethrow.append(ex0)

        # attempt to rebuild XREF from document
        if most_recent_xref is None:
            try:
                most_recent_xref = RebuiltXREF()
                assert most_recent_xref is not None
                most_recent_xref.set_parent(doc)  # type: ignore [attr-defined]
                most_recent_xref.read(src, tok)
                if "XRef" in doc:
                    doc[Name("XRef")] = doc["XRef"].merge(most_recent_xref)
                else:
                    doc[Name("XRef")] = most_recent_xref
            except Exception as ex0:
                most_recent_xref = None
                exceptions_to_rethrow.append(ex0)

        # unable to read XREF
        # re-throw exceptions
        if most_recent_xref is None:
            for e in exceptions_to_rethrow:
                raise e

        # handle Prev, Previous
        assert most_recent_xref is not None
        prev = None
        if "Prev" in most_recent_xref["Trailer"]:
            prev = int(most_recent_xref["Trailer"]["Prev"])
        if "Previous" in most_recent_xref["Trailer"]:
            prev = int(most_recent_xref["Trailer"]["Previous"])
        if prev is not None:
            self._read_xref(context, initial_offset=prev)



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of Event is triggered right after the Canvas has processed a stroke-path instruction.
"""
from borb.pdf.canvas.canvas_graphics_state import CanvasGraphicsState
from borb.pdf.canvas.event.event_listener import Event
from borb.pdf.canvas.geometry.line_segment import LineSegment


class LineRenderEvent(Event):
    """
    This implementation of Event is triggered right after the Canvas has processed a stroke-path instruction.
    """

    def __init__(self, graphics_state: CanvasGraphicsState, line_segment: LineSegment):
        super(LineRenderEvent, self).__init__()
        self._graphics_state = graphics_state
        self._line_segment = line_segment

    def get_line_segment(self) -> LineSegment:
        """
        Get the LineSegment that was constructed through various path-painting operators
        """
        return self._line_segment



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A text annotation represents a “sticky note” attached to a point in the PDF document. When closed, the
annotation shall appear as an icon; when open, it shall display a pop-up window containing the text of the note
in a font and size chosen by the conforming reader. Text annotations shall not scale and rotate with the page;
they shall behave as if the NoZoom and NoRotate annotation flags (see Table 165) were always set. Table 172
shows the annotation dictionary entries specific to this type of annotation.
"""
import enum
import typing

from borb.io.read.types import Name, Boolean
from borb.pdf.canvas.color.color import Color
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.annotation.annotation import Annotation


class TextAnnotationIconType(enum.Enum):
    """
    This Enum represents all possible text annotation icon types
    """

    COMMENT = Name("Comment")
    HELP = Name("Help")
    INSERT = Name("Insert")
    KEY = Name("Key")
    NEW_PARAGRAPH = Name("NewParagraph")
    NOTE = Name("Note")
    PARAGRAPH = Name("Paragraph")


class TextAnnotation(Annotation):
    """
    A text annotation represents a “sticky note” attached to a point in the PDF document. When closed, the
    annotation shall appear as an icon; when open, it shall display a pop-up window containing the text of the note
    in a font and size chosen by the conforming reader. Text annotations shall not scale and rotate with the page;
    they shall behave as if the NoZoom and NoRotate annotation flags (see Table 165) were always set. Table 172
    shows the annotation dictionary entries specific to this type of annotation.
    """

    def __init__(
        self,
        bounding_box: Rectangle,
        contents: str,
        color: typing.Optional[Color] = None,
        text_annotation_icon: TextAnnotationIconType = TextAnnotationIconType.COMMENT,
        open: typing.Optional[bool] = None,
    ):
        super(TextAnnotation, self).__init__(
            bounding_box=bounding_box, contents=contents, color=color
        )

        # specific for text annotations
        self[Name("Subtype")] = Name("Text")

        # (Optional) A flag specifying whether the annotation shall initially be
        # displayed open. Default value: false (closed).
        if open is not None:
            self[Name("Open")] = Boolean(open)

        # (Optional) The name of an icon that shall be used in displaying the
        # annotation. Conforming readers shall provide predefined icon
        # appearances for at least the following standard names:
        # Comment, Key, Note, Help, NewParagraph, Paragraph, Insert
        # Additional names may be supported as well. Default value: Note.
        # The annotation dictionary’s AP entry, if present, shall take precedence
        # over the Name entry; see Table 168 and 12.5.5, “Appearance Streams.”
        self[Name("Name")] = text_annotation_icon.value



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of LayoutElement represents a heterogeneous Paragraph.
e.g. a Paragraph where one or more words are in bold (but not all of them)
"""
import copy
import typing
from decimal import Decimal

from borb.pdf import Image
from borb.pdf.canvas.color.color import Color, HexColor
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.emoji.emoji import Emoji
from borb.pdf.canvas.layout.layout_element import Alignment
from borb.pdf.canvas.layout.text.chunk_of_text import ChunkOfText
from borb.pdf.canvas.layout.text.line_of_text import LineOfText
from borb.pdf.canvas.layout.text.paragraph import Paragraph
from borb.pdf.page.page import Page


class LineBreakChunk(ChunkOfText):
    """
    This implementation of ChunkOfText represents a linebreak.
    This can be used in the HeterogeneousParagraph to force lines to be split
    """

    def __init__(self):
        super(LineBreakChunk, self).__init__(text="\n")


class HeterogeneousParagraph(Paragraph):
    """
    This implementation of LayoutElement represents a heterogeneous collection of ChunkOfText elements.
    Furthermore (like Paragraph), this element has a default top- and bottom margin.
    e.g. a Paragraph where one or more words are in bold (but not all of them)
    """

    def __init__(
        self,
        chunks_of_text: typing.List[
            typing.Union[ChunkOfText, LineOfText, Emoji, Image, str]
        ] = [],
        vertical_alignment: Alignment = Alignment.TOP,
        horizontal_alignment: Alignment = Alignment.LEFT,
        text_alignment: Alignment = Alignment.LEFT,
        border_top: bool = False,
        border_right: bool = False,
        border_bottom: bool = False,
        border_left: bool = False,
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_bottom_left: Decimal = Decimal(0),
        border_color: Color = HexColor("000000"),
        border_width: Decimal = Decimal(1),
        padding_top: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        margin_top: typing.Optional[Decimal] = None,
        margin_right: typing.Optional[Decimal] = None,
        margin_bottom: typing.Optional[Decimal] = None,
        margin_left: typing.Optional[Decimal] = None,
        fixed_leading: typing.Optional[Decimal] = None,
        multiplied_leading: typing.Optional[Decimal] = None,
        background_color: typing.Optional[Color] = None,
    ):
        super(HeterogeneousParagraph, self).__init__(
            text="",
            text_alignment=text_alignment,
            vertical_alignment=vertical_alignment,
            horizontal_alignment=horizontal_alignment,
            border_top=border_top,
            border_right=border_right,
            border_bottom=border_bottom,
            border_left=border_left,
            border_radius_top_left=border_radius_top_left,
            border_radius_top_right=border_radius_top_right,
            border_radius_bottom_right=border_radius_bottom_right,
            border_radius_bottom_left=border_radius_bottom_left,
            border_color=border_color,
            border_width=border_width,
            padding_top=padding_top,
            padding_right=padding_right,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            margin_top=margin_top,
            margin_right=margin_right,
            margin_bottom=margin_bottom,
            margin_left=margin_left,
            fixed_leading=fixed_leading,
            multiplied_leading=multiplied_leading,
            background_color=background_color,
        )
        # fmt: off
        self._chunks_of_text: typing.List[typing.Union[ChunkOfText, LineOfText, Emoji, Image, str]] = chunks_of_text
        # fmt: on

    def _split_to_lines_of_chunks_of_text(
        self, available_space: Rectangle
    ) -> typing.List[typing.List[typing.Union[ChunkOfText, Emoji, Image]]]:

        # build list
        initial_chunks_of_text: typing.List[
            typing.Union[ChunkOfText, Emoji, Image]
        ] = []
        for e in self._chunks_of_text:
            if isinstance(e, str):
                initial_chunks_of_text.append(
                    ChunkOfText(
                        e,
                        border_top=self._border_top,
                        border_right=self._border_right,
                        border_bottom=self._border_bottom,
                        border_left=self._border_left,
                        border_radius_top_left=self._border_radius_top_left,
                        border_radius_top_right=self._border_radius_top_right,
                        border_radius_bottom_right=self._border_radius_bottom_right,
                        border_radius_bottom_left=self._border_radius_bottom_left,
                        background_color=self._background_color,
                        font_color=self._font_color,
                        font_size=self._font_size or Decimal(12),
                        font=self._font,
                    )
                )
                continue
            if isinstance(e, Emoji):
                initial_chunks_of_text.append(e)
                continue
            if isinstance(e, Image):
                initial_chunks_of_text.append(e)
                continue
            if isinstance(e, LineOfText):
                for word in e.get_text().split(" "):
                    # TODO: continuation of borders when splitting LineOfText in ChunkOfText
                    initial_chunks_of_text.append(
                        ChunkOfText(
                            word + " ",
                            border_top=e._border_top,
                            border_right=e._border_right,
                            border_bottom=e._border_bottom,
                            border_left=e._border_left,
                            border_radius_top_left=e._border_radius_top_left,
                            border_radius_top_right=e._border_radius_top_right,
                            border_radius_bottom_left=e._border_radius_bottom_left,
                            border_radius_bottom_right=e._border_radius_bottom_right,
                            background_color=e._background_color,
                            font_color=e._font_color,
                            font_size=e._font_size or Decimal(12),
                            font=e._font,
                        )
                    )
                continue
            if isinstance(e, ChunkOfText):
                initial_chunks_of_text.append(e)

        # perform initial layout (figure out where to break lines)
        next_x: Decimal = available_space.get_x()
        line_width: Decimal = Decimal(0)
        line: typing.List[typing.Union[ChunkOfText, Emoji, Image]] = []
        lines: typing.List[typing.List[typing.Union[ChunkOfText, Emoji, Image]]] = []
        for e in initial_chunks_of_text:
            w: Decimal = e.get_layout_box(
                Rectangle(
                    next_x,
                    available_space.get_y(),
                    available_space.get_width() - line_width,
                    available_space.get_height(),
                )
            ).get_width()
            if (line_width + w > available_space.get_width()) or (
                isinstance(e, LineBreakChunk)
            ):
                # get ready for next line
                lines.append(copy.deepcopy(line))
                next_x = available_space.get_x() + w
                line_width = w
                # fix the one element that does not fit
                assert e._previous_layout_box is not None
                e._previous_layout_box.x = available_space.get_x()
                line = [e]
                continue
            else:
                next_x += w
                line_width += w
                line.append(e)
        if len(line) > 0:
            lines.append(copy.deepcopy(line))

        # update ys
        prev_y: Decimal = available_space.get_y() + available_space.get_height()
        for line in lines:
            line_height: Decimal = max(
                [x.get_previous_layout_box().get_height() for x in line]
            )
            y: Decimal = prev_y - line_height
            for e in line:
                delta_height: Decimal = (
                    line_height - e.get_previous_layout_box().get_height()
                )
                e.get_layout_box(
                    Rectangle(
                        e.get_previous_layout_box().get_x(),
                        y,
                        e.get_previous_layout_box().get_width(),
                        line_height - delta_height,
                    )
                )
            prev_y -= line_height

        # update xs (text_alignment == CENTERED)
        delta_x: Decimal = Decimal(0)
        line_max_x: Decimal = Decimal(0)
        cbox_max_x: Decimal = Decimal(0)
        if self._text_alignment == Alignment.CENTERED:
            for line in lines:
                line_max_x = (
                    line[-1].get_previous_layout_box().get_x()
                    + line[-1].get_previous_layout_box().get_width()
                )
                cbox_max_x = available_space.get_x() + available_space.get_width()
                delta_x = (cbox_max_x - line_max_x) / 2
                for e in line:
                    e._previous_layout_box.x += delta_x

        # update xs (text_alignment == LEFT)
        if self._text_alignment == Alignment.RIGHT:
            for line in lines:
                line_max_x = (
                    line[-1].get_previous_layout_box().get_x()
                    + line[-1].get_previous_layout_box().get_width()
                )
                cbox_max_x = available_space.get_x() + available_space.get_width()
                delta_x = cbox_max_x - line_max_x
                for e in line:
                    e._previous_layout_box.x += delta_x

        # update xs (text_alignment == JUSTIFIED)
        if self._text_alignment == Alignment.JUSTIFIED:
            for line in lines:
                line_max_x = (
                    line[-1].get_previous_layout_box().get_x()
                    + line[-1].get_previous_layout_box().get_width()
                )
                cbox_max_x = available_space.get_x() + available_space.get_width()
                delta_x = (cbox_max_x - line_max_x) / (len(line) - 1)
                for e in line:
                    e._previous_layout_box.x += delta_x

        # return
        return lines

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:

        # lines of text
        lines_of_text: typing.List[
            typing.List[ChunkOfText]
        ] = self._split_to_lines_of_chunks_of_text(available_space)

        # determine height
        h: Decimal = sum(
            [
                max([x.get_previous_layout_box().get_height() for x in line_of_text])
                for line_of_text in lines_of_text
            ]
        )

        # determine width
        w: Decimal = max(
            [
                sum([x.get_previous_layout_box().get_width() for x in line_of_text])
                for line_of_text in lines_of_text
            ]
        )
        if self._text_alignment == Alignment.JUSTIFIED:
            w = available_space.get_width()

        # return
        return Rectangle(
            available_space.get_x(),
            available_space.get_y() + available_space.get_height() - h,
            w,
            h,
        )

    def _paint_content_box(self, page: Page, available_space: Rectangle) -> None:

        # lines of text
        lines_of_text: typing.List[
            typing.List[ChunkOfText]
        ] = self._split_to_lines_of_chunks_of_text(available_space)

        for line in lines_of_text:
            for e in line:
                lbox: typing.Optional[Rectangle] = e.get_previous_layout_box()
                assert lbox is not None
                e.paint(page, lbox)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This module contains common stopwords in French and English
"""

# fmt: off
ENGLISH_STOP_WORDS = [
    "0o", "0s", "3a", "3b", "3d", "6b", "6o", "a", "a1", "a2", "a3", "a4", "ab", "able", "about", "above", "abst", "ac", "accordance", "according", "accordingly", "across", "act", "actually", "ad", "added", "adj", "ae", "af", "affected", "affecting", "affects", "after", "afterwards", "ag", "again", "against", "ah", "ain", "ain't", "aj", "al", "all", "allow", "allows", "almost", "alone", "along", "already", "also", "although", "always", "am", "among", "amongst", "amoungst", "amount", "an", "and", "announce", "another", "any", "anybody", "anyhow",
    "anymore", "anyone", "anything", "anyway", "anyways", "anywhere", "ao", "ap", "apart", "apparently", "appear", "appreciate", "appropriate", "approximately", "ar", "are", "aren", "arent", "aren't", "arise", "around", "as", "a's", "aside", "ask", "asking", "associated", "at", "au", "auth", "av", "available", "aw", "away", "awfully", "ax", "ay", "az", "b", "b1", "b2", "b3", "ba", "back", "bc", "bd", "be", "became", "because", "become", "becomes", "becoming", "been", "before", "beforehand", "begin", "beginning", "beginnings", "begins", "behind", "being", "believe", "below", "beside",
    "besides", "best", "better", "between", "beyond", "bi", "bill", "biol", "bj", "bk", "bl", "bn", "both", "bottom", "bp", "br", "brief", "briefly", "bs", "bt", "bu", "but", "bx", "by", "c", "c1", "c2", "c3", "ca", "call", "came", "can", "cannot", "cant", "can't", "cause", "causes", "cc", "cd", "ce", "certain", "certainly", "cf", "cg", "ch", "changes", "ci", "cit", "cj", "cl", "clearly", "cm", "c'mon", "cn", "co", "com", "come", "comes", "con", "concerning", "consequently", "consider", "considering", "contain",
    "containing", "contains", "corresponding", "could", "couldn", "couldnt", "couldn't", "course", "cp", "cq", "cr", "cry", "cs", "c's", "ct", "cu", "currently", "cv", "cx", "cy", "cz", "d", "d2", "da", "date", "dc", "dd", "de", "definitely", "describe", "described", "despite", "detail", "df", "di", "did", "didn", "didn't", "different", "dj", "dk", "dl", "do", "does", "doesn", "doesn't", "doing", "don", "done", "don't", "down", "downwards", "dp", "dr", "ds", "dt", "du", "due", "during", "dx", "dy", "e", "e2", "e3",
    "ea", "each", "ec", "ed", "edu", "ee", "ef", "effect", "eg", "ei", "eight", "eighty", "either", "ej", "el", "eleven", "else", "elsewhere", "em", "empty", "en", "end", "ending", "enough", "entirely", "eo", "ep", "eq", "er", "es", "especially", "est", "et", "et-al", "etc", "eu", "ev", "even", "ever", "every", "everybody", "everyone", "everything", "everywhere", "ex", "exactly", "example", "except", "ey", "f", "f2", "fa", "far", "fc", "few", "ff", "fi", "fifteen", "fifth", "fify", "fill", "find", "fire", "first",
    "five", "fix", "fj", "fl", "fn", "fo", "followed", "following", "follows", "for", "former", "formerly", "forth", "forty", "found", "four", "fr", "from", "front", "fs", "ft", "fu", "full", "further", "furthermore", "fy", "g", "ga", "gave", "ge", "get", "gets", "getting", "gi", "give", "given", "gives", "giving", "gj", "gl", "go", "goes", "going", "gone", "got", "gotten", "gr", "greetings", "gs", "gy", "h", "h2", "h3", "had", "hadn", "hadn't", "happens", "hardly", "has", "hasn", "hasnt", "hasn't", "have", "haven",
    "haven't", "having", "he", "hed", "he'd", "he'll", "hello", "help", "hence", "her", "here", "hereafter", "hereby", "herein", "heres", "here's", "hereupon", "hers", "herself", "hes", "he's", "hh", "hi", "hid", "him", "himself", "his", "hither", "hj", "ho", "home", "hopefully", "how", "howbeit", "however", "how's", "hr", "hs", "http", "hu", "hundred", "hy", "i", "i2", "i3", "i4", "i6", "i7", "i8", "ia", "ib", "ibid", "ic", "id", "i'd", "ie", "if", "ig", "ignored", "ih", "ii", "ij", "il", "i'll",
    "im", "i'm", "immediate", "immediately", "importance", "important", "in", "inasmuch", "inc", "indeed", "index", "indicate", "indicated", "indicates", "information", "inner", "insofar", "instead", "interest", "into", "invention", "inward", "io", "ip", "iq", "ir", "is", "isn", "isn't", "it", "itd", "it'd", "it'll", "its", "it's", "itself", "iv", "i've", "ix", "iy", "iz", "j", "jj", "jr", "js", "jt", "ju", "just", "k", "ke", "keep", "keeps", "kept", "kg", "kj", "km", "know", "known", "knows", "ko", "l", "l2", "la", "largely",
    "last", "lately", "later", "latter", "latterly", "lb", "lc", "le", "least", "les", "less", "lest", "let", "lets", "let's", "lf", "like", "liked", "likely", "line", "little", "lj", "ll", "ll", "ln", "lo", "look", "looking", "looks", "los", "lr", "ls", "lt", "ltd", "m", "m2", "ma", "made", "mainly", "make", "makes", "many", "may", "maybe", "me", "mean", "means", "meantime", "meanwhile", "merely", "mg", "might", "mightn", "mightn't", "mill", "million", "mine", "miss", "ml", "mn", "mo", "more", "moreover", "most",
    "mostly", "move", "mr", "mrs", "ms", "mt", "mu", "much", "mug", "must", "mustn", "mustn't", "my", "myself", "n", "n2", "na", "name", "namely", "nay", "nc", "nd", "ne", "near", "nearly", "necessarily", "necessary", "need", "needn", "needn't", "needs", "neither", "never", "nevertheless", "new", "next", "ng", "ni", "nine", "ninety", "nj", "nl", "nn", "no", "nobody", "non", "none", "nonetheless", "noone", "nor", "normally", "nos", "not", "noted", "nothing", "novel", "now", "nowhere", "nr", "ns", "nt", "ny", "o", "oa",
    "ob", "obtain", "obtained", "obviously", "oc", "od", "of", "off", "often", "og", "oh", "oi", "oj", "ok", "okay", "ol", "old", "om", "omitted", "on", "once", "one", "ones", "only", "onto", "oo", "op", "oq", "or", "ord", "os", "ot", "other", "others", "otherwise", "ou", "ought", "our", "ours", "ourselves", "out", "outside", "over", "overall", "ow", "owing", "own", "ox", "oz", "p", "p1", "p2", "p3", "page", "pagecount", "pages", "par", "part", "particular", "particularly", "pas", "past", "pc", "pd",
    "pe", "per", "perhaps", "pf", "ph", "pi", "pj", "pk", "pl", "placed", "please", "plus", "pm", "pn", "po", "poorly", "possible", "possibly", "potentially", "pp", "pq", "pr", "predominantly", "present", "presumably", "previously", "primarily", "probably", "promptly", "proud", "provides", "ps", "pt", "pu", "put", "py", "q", "qj", "qu", "que", "quickly", "quite", "qv", "r", "r2", "ra", "ran", "rather", "rc", "rd", "re", "readily", "really", "reasonably", "recent", "recently", "ref", "refs", "regarding", "regardless", "regards", "related", "relatively", "research",
    "research-articl", "respectively", "resulted", "resulting", "results", "rf", "rh", "ri", "right", "rj", "rl", "rm", "rn", "ro", "rq", "rr", "rs", "rt", "ru", "run", "rv", "ry", "s", "s2", "sa", "said", "same", "saw", "say", "saying", "says", "sc", "sd", "se", "sec", "second", "secondly", "section", "see", "seeing", "seem", "seemed", "seeming", "seems", "seen", "self", "selves", "sensible", "sent", "serious", "seriously", "seven", "several", "sf", "shall", "shan", "shan't", "she", "shed", "she'd", "she'll", "shes", "she's", "should",
    "shouldn", "shouldn't", "should've", "show", "showed", "shown", "showns", "shows", "si", "side", "significant", "significantly", "similar", "similarly", "since", "sincere", "six", "sixty", "sj", "sl", "slightly", "sm", "sn", "so", "some", "somebody", "somehow", "someone", "somethan", "something", "sometime", "sometimes", "somewhat", "somewhere", "soon", "sorry", "sp", "specifically", "specified", "specify", "specifying", "sq", "sr", "ss", "st", "still", "stop", "strongly", "sub", "substantially", "successfully", "such", "sufficiently", "suggest", "sup", "sure", "sy", "system", "sz", "t", "t1", "t2", "t3", "take",
    "taken", "taking", "tb", "tc", "td", "te", "tell", "ten", "tends", "tf", "th", "than", "thank", "thanks", "thanx", "that", "that'll", "thats", "that's", "that've", "the", "their", "theirs", "them", "themselves", "then", "thence", "there", "thereafter", "thereby", "thered", "therefore", "therein", "there'll", "thereof", "therere", "theres", "there's", "thereto", "thereupon", "there've", "these", "they", "theyd", "they'd", "they'll", "theyre", "they're", "they've", "thickv", "thin", "think", "third", "this", "thorough", "thoroughly", "those", "thou", "though", "thoughh", "thousand", "three", "throug", "through",
    "throughout", "thru", "thus", "ti", "til", "tip", "tj", "tl", "tm", "tn", "to", "together", "too", "took", "top", "toward", "towards", "tp", "tq", "tr", "tried", "tries", "truly", "try", "trying", "ts", "t's", "tt", "tv", "twelve", "twenty", "twice", "two", "tx", "u", "u201d", "ue", "ui", "uj", "uk", "um", "un", "under", "unfortunately", "unless", "unlike", "unlikely", "until", "unto", "uo", "up", "upon", "ups", "ur", "us", "use", "used", "useful", "usefully", "usefulness", "uses", "using", "usually", "ut",
    "v", "va", "value", "various", "vd", "ve", "ve", "very", "via", "viz", "vj", "vo", "vol", "vols", "volumtype", "vq", "vs", "vt", "vu", "w", "wa", "want", "wants", "was", "wasn", "wasnt", "wasn't", "way", "we", "wed", "we'd", "welcome", "well", "we'll", "well-b", "went", "were", "we're", "weren", "werent", "weren't", "we've", "what", "whatever", "what'll", "whats", "what's", "when", "whence", "whenever", "when's", "where", "whereafter", "whereas", "whereby", "wherein", "wheres", "where's", "whereupon", "wherever", "whether", "which", "while", "whim",
    "whither", "who", "whod", "whoever", "whole", "who'll", "whom", "whomever", "whos", "who's", "whose", "why", "why's", "wi", "widely", "will", "willing", "wish", "with", "within", "without", "wo", "won", "wonder", "wont", "won't", "words", "world", "would", "wouldn", "wouldnt", "wouldn't", "www", "x", "x1", "x2", "x3", "xf", "xi", "xj", "xk", "xl", "xn", "xo", "xs", "xt", "xv", "xx", "y", "y2", "yes", "yet", "yj", "yl", "you", "youd", "you'd", "you'll", "your", "youre", "you're", "yours", "yourself", "yourselves",
    "you've", "yr", "ys", "yt", "z", "zero", "zi", "zz"
]
# fmt: on

# fmt: off
FRENCH_STOP_WORDS = [
    "alors", "au", "aucuns", "aussi", "autre", "avant", "avec", "avoir", "bon", "car", "ce", "cela", "ces", "ceux", "chaque", "ci", "comme", "comment", "dans", "des", "du", "dedans", "dehors", "depuis", "devrait", "doit", "donc", "dos", "début", "elle", "elles", "en", "encore", "essai", "est", "et", "eu", "fait", "faites", "fois", "_font", "hors", "ici", "il", "ils", "je	", "juste", "la", "le", "les", "leur", "là", "ma", "maintenant", "mais", "mes", "mien", "moins", "mon", "mot", "même", "ni", "nommés", "notre",
    "nous", "ou", "où", "par", "parce", "pas", "peut", "peu", "plupart", "pour", "pourquoi", "quand", "que", "quel", "quelle", "quelles", "quels", "qui", "sa", "sans", "ses", "seulement", "si", "sien", "son", "sont", "sous", "soyez	", "sujet", "sur", "ta", "tandis", "tellement", "tels", "tes", "ton", "tous", "tout", "trop", "très", "tu", "voient", "vont", "votre", "vous", "vu", "ça", "étaient", "état", "étions", "été", "être"
]
# fmt: on



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Close and stroke the path. This operator shall have the same effect as the
sequence h S.
"""
import typing

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class CloseAndStrokePath(CanvasOperator):
    """
    Close and stroke the path. This operator shall have the same effect as the
    sequence h S.
    """

    def __init__(self):
        super().__init__("s", 0)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the s operator
        """
        close_subpath_op: typing.Optional[
            CanvasOperator
        ] = canvas_stream_processor.get_operator("h")
        assert close_subpath_op
        close_subpath_op.invoke(canvas_stream_processor, [], event_listeners)

        stroke_path_op: typing.Optional[
            CanvasOperator
        ] = canvas_stream_processor.get_operator("S")
        assert stroke_path_op
        stroke_path_op.invoke(canvas_stream_processor, [], event_listeners)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of WriteBaseTransformer is responsible for writing XREF objects
"""
import typing
from typing import Optional

from borb.io.read.types import AnyPDFType, Decimal, Dictionary, Name, Reference
from borb.io.write.transformer import Transformer, WriteTransformerState
from borb.pdf.xref.xref import XREF


class XREFTransformer(Transformer):
    """
    This implementation of WriteBaseTransformer is responsible for writing XREF objects
    """

    def can_be_transformed(self, any: AnyPDFType):
        """
        This function returns True if the object to be converted represents a cross-reference table
        """
        return isinstance(any, XREF)

    def transform(
        self,
        object_to_transform: AnyPDFType,
        context: Optional[WriteTransformerState] = None,
    ):
        """
        This method writes an XREF to a byte stream
        """
        assert isinstance(object_to_transform, XREF)
        assert "Trailer" in object_to_transform
        assert isinstance(object_to_transform["Trailer"], Dictionary)
        # fmt: off
        assert context is not None, "A WriteTransformerState must be defined in order to write XREF objects."
        assert context.destination is not None, "A WriteTransformerState must be defined in order to write XREF objects."
        # fmt: on

        # Transform the Trailer dictionary (replacing objects by references)
        # we do this upfront because the normal write_dictionary_transformer will write the dictionary first,
        # and the references afterwards. This would cause the /Trailer dictionary to not be the last.
        trailer_out = Dictionary()

        # /Root
        trailer_out[Name("Root")] = self.get_reference(
            object_to_transform["Trailer"]["Root"], context
        )

        # /Info
        if "Info" in object_to_transform["Trailer"]:
            trailer_out[Name("Info")] = self.get_reference(
                object_to_transform["Trailer"]["Info"], context
            )

        # /Size
        if (
            "Trailer" in object_to_transform
            and "Size" in object_to_transform["Trailer"]
        ):
            trailer_out[Name("Size")] = object_to_transform["Trailer"]["Size"]
        else:
            trailer_out[Name("Size")] = Decimal(
                0
            )  # we'll recalculate this later anyway

        # /ID
        if "ID" in object_to_transform["Trailer"]:
            trailer_out[Name("ID")] = object_to_transform["Trailer"]["ID"]

        # write /Info object
        if "Info" in object_to_transform["Trailer"]:
            self.get_root_transformer().transform(
                object_to_transform["Trailer"]["Info"], context
            )

        # write /Root object
        self.get_root_transformer().transform(
            object_to_transform["Trailer"]["Root"], context
        )

        # write /XREF
        start_of_xref = context.destination.tell()
        context.destination.write(bytes("xref\n", "latin1"))
        for section in self._section_xref(context):
            context.destination.write(
                bytes("%d %d\n" % (section[0].object_number, len(section)), "latin1")
            )
            for r in section:
                if r.is_in_use:
                    context.destination.write(
                        bytes("{0:010d} 00000 n\r\n".format(r.byte_offset), "latin1")
                    )
                else:
                    context.destination.write(
                        bytes("{0:010d} 00000 f\r\n".format(r.byte_offset), "latin1")
                    )

        # update /Size
        trailer_out[Name("Size")] = Decimal(
            sum([len(v) for k, v in context.indirect_objects_by_hash.items()]) + 1
        )

        # write /Trailer
        context.destination.write(bytes("trailer\n", "latin1"))
        self.get_root_transformer().transform(trailer_out, context)
        context.destination.write(bytes("startxref\n", "latin1"))

        # write byte offset of last cross-reference section
        context.destination.write(bytes(str(start_of_xref) + "\n", "latin1"))

        # write EOF
        context.destination.write(bytes("%%EOF", "latin1"))

    def _section_xref(self, context: Optional[WriteTransformerState] = None):
        assert (
            context is not None
        ), "A WriteTransformerState must be defined in order to write XREF objects."

        # get all references
        indirect_objects: typing.List[AnyPDFType] = [
            item
            for sublist in [v for k, v in context.indirect_objects_by_hash.items()]
            for item in sublist
        ]
        references: typing.List[Reference] = []
        for obj in indirect_objects:
            ref = obj.get_reference()  # type: ignore [union-attr]
            if ref is not None:
                references.append(ref)
        # sort
        references.sort(key=lambda x: x.object_number)

        # insert magic entry if needed
        if len(references) == 0 or references[0].generation_number != 65535:
            references.insert(
                0,
                Reference(
                    generation_number=65535,
                    object_number=0,
                    byte_offset=0,
                    is_in_use=False,
                ),
            )

        # divide into sections
        sections = [[references[0]]]
        for i in range(1, len(references)):
            ref = references[i]
            prev_object_number = sections[-1][-1].object_number
            assert prev_object_number is not None
            if ref.object_number == prev_object_number + 1:
                sections[-1].append(ref)
            else:
                sections.append([ref])

        # return
        return sections



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of ReadBaseTransformer is responsible for reading String objects
"""
import io
import typing
from typing import Any, Optional, Union

from borb.io.read.transformer import ReadTransformerState, Transformer
from borb.io.read.types import AnyPDFType, HexadecimalString, Name, String
from borb.pdf.canvas.event.event_listener import EventListener


class StringTransformer(Transformer):
    """
    This implementation of ReadBaseTransformer is responsible for reading String objects
    """

    def can_be_transformed(
        self, object: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO, AnyPDFType]
    ) -> bool:
        """
        This function returns True if the object to be converted represents a String or Hexadecimal String or a Name
        """
        return (
            isinstance(object, String)
            or isinstance(object, HexadecimalString)
            or isinstance(object, Name)
        )

    def transform(
        self,
        object_to_transform: Union[io.BufferedIOBase, io.RawIOBase, AnyPDFType],
        parent_object: Any,
        context: Optional[ReadTransformerState] = None,
        event_listeners: typing.List[EventListener] = [],
    ) -> Any:
        """
        This function reads a String from a byte stream
        """
        # set parent
        object_to_transform.set_parent(parent_object)  # type: ignore[union-attr]
        # return
        return object_to_transform



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Restore the graphics state by removing the most recently saved
state from the stack and making it the current state (see 8.4.2,
"Graphics State Stack").
"""

import typing

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class PopGraphicsState(CanvasOperator):
    """
    Restore the graphics state by removing the most recently saved
    state from the stack and making it the current state (see 8.4.2,
    "Graphics State Stack").
    """

    def __init__(self):
        super().__init__("Q", 0)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the Q operator
        """
        canvas = canvas_stream_processor.get_canvas()
        assert (
            len(canvas.graphics_state_stack) > 0
        ), "Stack underflow. Q operator was applied to an empty stack."
        canvas.graphics_state = canvas.graphics_state_stack.pop(-1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
In Euclidean plane geometry, a rectangle is a quadrilateral with four right angles.
"""
from decimal import Decimal


class Rectangle:
    """
    In Euclidean plane geometry, a rectangle is a quadrilateral with four right angles.
    It can also be defined as an equiangular quadrilateral, since equiangular means that all of its angles are equal (360°/4 = 90°).
    It can also be defined as a parallelogram containing a right angle. A rectangle with four sides of equal length is a square.
    The term oblong is occasionally used to refer to a non-square rectangle.
    """

    def __init__(
        self,
        lower_left_x: Decimal,
        lower_left_y: Decimal,
        width: Decimal,
        height: Decimal,
    ):
        assert width >= 0, "A Rectangle must have a non-negative width."
        assert height >= 0, "A Rectangle must have a non-negative height."
        self.x = lower_left_x
        self.y = lower_left_y
        self.width = width
        self.height = height

    def get_x(self) -> Decimal:
        """
        This function returns the x-coordinate of the lower-left of this Rectangle
        """
        return self.x

    def get_y(self) -> Decimal:
        """
        This function returns the y-coordinate of the lower-left of this Rectangle
        """
        return self.y

    def get_width(self) -> Decimal:
        """
        This function returns the width of this Rectangle
        """
        return self.width

    def get_height(self) -> Decimal:
        """
        This function returns the height of this Rectangle
        """
        return self.height

    def intersects(self, other_rectangle: "Rectangle") -> bool:
        """
        This function returns True if this Rectangle intersects with the given Rectangle,
        False otherwise
        """
        # fmt: off
        x_intersect: bool = (other_rectangle.x <= self.x <= (other_rectangle.x + other_rectangle.width))
        x_intersect = x_intersect or (self.x <= other_rectangle.x <= (self.x + self.width))
        y_intersect: bool = (other_rectangle.y <= self.y <= (other_rectangle.y + other_rectangle.height))
        y_intersect = y_intersect or (self.y <= other_rectangle.y <= (self.y + self.height))
        # fmt: on
        return x_intersect and y_intersect

    def contains(self, x: Decimal, y: Decimal) -> bool:
        """
        This function returns True if this Rectangle contains the given point
        False otherwise
        """
        return self.x <= x <= (self.x + self.width) and self.y <= y <= (
            self.y + self.height
        )

    def circumference_contains(self, x: Decimal, y: Decimal) -> bool:
        """
        This function returns True if the circumference of this Rectangle contains the given point
        False otherwise
        """
        return x in [self.x, self.x + self.width] and y in [
            self.y,
            self.y + self.height,
        ]

    def shrink(self, amount: Decimal) -> "Rectangle":
        """
        This function returns a (slightly) smaller Rectangle,
        shrunk by the given amount
        """
        assert amount >= 0
        return Rectangle(
            self.x + amount,
            self.y + amount,
            self.width - Decimal(2) * amount,
            self.height - Decimal(2) * amount,
        )

    def grow(self, amount: Decimal) -> "Rectangle":
        """
        This function returns a (slightly) larger Rectangle,
        grown by the given amount
        """
        assert amount >= 0
        return Rectangle(
            self.x - amount,
            self.y - amount,
            self.width + Decimal(2) * amount,
            self.height + Decimal(2) * amount,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of ReadBaseTransformer is responsible for reading a grayscale image object
"""
import io
import logging
import typing
from typing import Any, Optional, Union

from PIL import Image  # type: ignore [import]

from borb.io.filter.stream_decode_util import decode_stream
from borb.io.read.transformer import ReadTransformerState, Transformer
from borb.io.read.types import AnyPDFType, Reference, Stream, add_base_methods
from borb.pdf.canvas.event.event_listener import EventListener

logger = logging.getLogger(__name__)


class GrayscaleImageTransformer(Transformer):
    """
    This implementation of ReadBaseTransformer is responsible for reading a grayscale image object
    """

    def can_be_transformed(
        self, object: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO, AnyPDFType]
    ) -> bool:
        """
        This function returns True if the object to be transformed is a grayscale Image
        """
        return (
            isinstance(object, Stream)
            and object.get("Type", None) in ["XObject", None]
            and object.get("Subtype", None) == ("Image")
            and "Filter" in object
            and (
                object["Filter"] == "FlateDecode"
                or (
                    isinstance(object["Filter"], list)
                    and object["Filter"][0] == "FlateDecode"
                )
            )
            and object.get("ColorSpace", None) == "DeviceGray"
        )

    def transform(
        self,
        object_to_transform: Union["io.IOBase", AnyPDFType],
        parent_object: Any,
        context: Optional[ReadTransformerState] = None,
        event_listeners: typing.List[EventListener] = [],
    ) -> Any:
        """
        This function reads a grayscale Image from a byte stream
        """

        # fmt: off
        assert isinstance(object_to_transform, Stream), "object_to_transform must be of type Stream"
        # fmt: on

        # resolve references in stream dictionary
        xref = parent_object.get_root().get("XRef")
        for k, v in object_to_transform.items():
            if isinstance(v, Reference):
                # fmt: off
                assert (context is not None), "context must be defined to read Image objects"
                assert (context.source is not None), "context.source must be defined to read Image objects"
                assert (context.tokenizer is not None), "context.tokenizer must be defined to read Image objects"
                # fmt: on
                v = xref.get_object(v, context.source, context.tokenizer)
                object_to_transform[k] = v

        grayscale_bytes = [
            x for x in decode_stream(object_to_transform)["DecodedBytes"]
        ]

        # use PIL to process image bytes
        w = int(object_to_transform["Width"])
        h = int(object_to_transform["Height"])
        tmp = Image.new("RGB", (w, h))
        for i in range(0, w):
            for j in range(0, h):
                k = i * h + j
                try:
                    c = (grayscale_bytes[k], grayscale_bytes[k], grayscale_bytes[k])
                    tmp.putpixel((i, j), value=c)
                except:
                    pass

        # add base methods
        add_base_methods(tmp)

        # set parent
        tmp.set_parent(parent_object)  # type: ignore[attr-defined]

        # return
        return tmp



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of WriteBaseTransformer is responsible for writing References
"""
from typing import Optional

from borb.io.read.types import AnyPDFType, Reference
from borb.io.write.transformer import Transformer, WriteTransformerState


class ReferenceTransform(Transformer):
    """
    This implementation of WriteBaseTransformer is responsible for writing References
    """

    def can_be_transformed(self, any: AnyPDFType):
        """
        This function returns True if the object to be converted represents a Reference
        """
        return isinstance(any, Reference)

    def transform(
        self,
        object_to_transform: AnyPDFType,
        context: Optional[WriteTransformerState] = None,
    ):
        """
        This method writes a Reference to a byte stream
        """
        assert (
            context is not None
        ), "A WriteTransformerState must be defined in order to write Reference objects."
        assert context.destination is not None
        assert isinstance(object_to_transform, Reference)

        assert object_to_transform.object_number is not None
        context.destination.write(
            bytes(
                "%d %d R"
                % (
                    object_to_transform.object_number,
                    object_to_transform.generation_number or 0,
                ),
                "latin1",
            )
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
An operator in a programming language is a symbol that tells the compiler or interpreter to perform specific mathematical,
relational or logical operation and produce final result.
CanvasOperator defines an interface to work on Canvas objects. Typically these operators involve drawing graphics, text,
setting the active color and so on
"""
import typing

from borb.io.read.types import AnyPDFType


class CanvasOperator:
    """
    An operator in a programming language is a symbol that tells the compiler or interpreter to perform specific mathematical,
    relational or logical operation and produce final result.
    CanvasOperator defines an interface to work on Canvas objects. Typically these operators involve drawing graphics, text,
    setting the active color and so on
    """

    def __init__(self, text: str, number_of_operands: int):
        self._text: str = text
        self._number_of_operands: int = number_of_operands

    def get_text(self) -> str:
        """
        Return the str that invokes this CanvasOperator
        """
        return self._text

    def get_number_of_operands(self) -> int:
        """
        Return the number of operands for this CanvasOperator
        """
        return self._number_of_operands

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invokes this CanvasOperator
        """
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of WriteBaseTransformer is responsible for writing /Info Dictionary objects
"""
import datetime
import logging
import random
import typing
import xml.etree.ElementTree as ET
from typing import Any, Optional

from borb.io.read.types import AnyPDFType
from borb.io.read.types import Decimal as bDecimal
from borb.io.read.types import Dictionary, Name, Stream, String
from borb.io.write.conformance_level import ConformanceLevel
from borb.io.write.object.dictionary_transformer import DictionaryTransformer
from borb.io.write.object.stream_transformer import StreamTransformer
from borb.io.write.transformer import Transformer, WriteTransformerState
from borb.pdf.document.document import Document
from borb.pdf.trailer.document_info import XMPDocumentInfo

logger = logging.getLogger(__name__)


class InformationDictionaryTransformer(Transformer):
    """
    This implementation of WriteBaseTransformer is responsible for writing /Info Dictionary objects
    """

    def can_be_transformed(self, any: AnyPDFType):
        """
        This function returns True if the object to be transformed is an /Info Dictionary
        """
        if not isinstance(any, Dictionary):
            return False
        parent: AnyPDFType = any.get_parent()  # type: ignore [attr-defined]
        return (
            isinstance(parent, Dictionary)
            and "Info" in parent
            and parent["Info"] == any
        )

    #
    # XMP
    #

    def _consolidate_xmp_and_info_dictionary(self, document: Document) -> Dictionary:
        new_info_dictionary: Dictionary = Dictionary()

        # get /Info Dictionary
        if (
            "XRef" in document
            and "Trailer" in document["XRef"]
            and "Info" in document["XRef"]["Trailer"]
            and isinstance(document["XRef"]["Trailer"]["Info"], Dictionary)
        ):
            info_dictionary: Dictionary = document["XRef"]["Trailer"]["Info"]
            for k, v in info_dictionary.items():
                new_info_dictionary[k] = v

        # get XMP /Metadata
        if (
            "XRef" in document
            and "Trailer" in document["XRef"]
            and "Root" in document["XRef"]["Trailer"]
            and "Metadata" in document["XRef"]["Trailer"]["Root"]
            and isinstance(document["XRef"]["Trailer"]["Root"]["Metadata"], ET.Element)
        ):
            xmp_document_info: XMPDocumentInfo = document.get_xmp_document_info()
            for k, v in {
                Name("Title"): xmp_document_info.get_title(),
                Name("Author"): xmp_document_info.get_author(),
                Name("Subject"): xmp_document_info.get_subject(),
                Name("Keywords"): xmp_document_info.get_keywords(),
                Name("Creator"): xmp_document_info.get_creator(),
                Name("Producer"): xmp_document_info.get_producer(),
                Name("CreationDate"): xmp_document_info.get_creation_date(),
                Name("ModDate"): xmp_document_info.get_modification_date(),
            }.items():
                if v is None:
                    continue
                if k in ["CreationDate", "ModDate"]:
                    v = InformationDictionaryTransformer._convert_xmp_date_format_to_iso_8824_date_format(
                        v
                    )
                new_info_dictionary[k] = String(v)

        # return
        return new_info_dictionary

    @staticmethod
    def _now_as_iso_8824_date_format() -> str:
        timestamp_str = "D:"
        now = datetime.datetime.now()
        for n in [now.year, now.month, now.day, now.hour, now.minute, now.second]:
            timestamp_str += "{0:02}".format(n)
        timestamp_str += "Z00"
        return timestamp_str

    def _update_info_dictionary(self, info_dictionary: Dictionary) -> Dictionary:

        # set CreationDate
        if "CreationDate" not in info_dictionary:
            info_dictionary[Name("CreationDate")] = String(
                InformationDictionaryTransformer._now_as_iso_8824_date_format()
            )

        # set ModDate
        info_dictionary[Name("ModDate")] = String(
            InformationDictionaryTransformer._now_as_iso_8824_date_format()
        )

        # set Producer
        info_dictionary[Name("Producer")] = String("borb")

        return info_dictionary

    @staticmethod
    def _convert_iso_8824_date_format_to_xmp_date_format(s: str) -> str:
        try:
            year: str = s[2:6]
            month: str = s[6:8]
            day: str = s[8:10]
            hour: str = s[10:12]
            minute: str = s[12:14]
            second: str = s[14:16]
            # fmt: off
            return year + "-" + month + "-" + day + "T" + hour + ":" + minute + ":" + second + "+00:00"
            # fmt: on
        except:
            return s

    @staticmethod
    def _convert_xmp_date_format_to_iso_8824_date_format(s: str) -> str:
        return s

    def _write_xmp_metadata_stream(
        self,
        info_dictionary: Dictionary,
        conformance_level: typing.Optional[ConformanceLevel] = None,
    ) -> Stream:
        random_id: str = "".join(
            [
                random.choice("0123456789abcdefghijklmnopqrstuvwxyz")
                for _ in range(0, 24)
            ]
        )
        s: str = '<?xpacket begin="" id="%s"?>' % random_id
        s += '\n<x:xmpmeta xmlns:x="adobe:ns:meta/" x:xmptk="Adobe XMP Core 5.1.0-jc003">'
        s += '\n\t<rdf:RDF xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#">'

        # rdf:Description
        d: typing.Dict[str, str] = {
            "rdf:about": "",
            "xmlns:dc": "http://purl.org/dc/elements/1.1/",
            "xmlns:pdf": "http://ns.adobe.com/pdf/1.3/",
            "xmlns:xmp": "http://ns.adobe.com/xap/1.0/",
            "dc:format": "application/pdf",
        }
        if "Producer" in info_dictionary:
            d["pdf:Producer"] = str(info_dictionary["Producer"])
        if "Keywords" in info_dictionary:
            d["pdf:Keywords"] = str(info_dictionary["Keywords"])
        if "CreationDate" in info_dictionary:
            # fmt: off
            d["xmp:CreateDate"] = InformationDictionaryTransformer._convert_iso_8824_date_format_to_xmp_date_format(info_dictionary["CreationDate"])
            # fmt: on
        if "Creator" in info_dictionary:
            d["xmp:CreatorTool"] = str(info_dictionary["Creator"])
        if "ModDate" in info_dictionary:
            # fmt: off
            d["xmp:ModifyDate"] = InformationDictionaryTransformer._convert_iso_8824_date_format_to_xmp_date_format(info_dictionary["ModDate"])
            # fmt: on
        s += (
            "\n\t\t<rdf:Description"
            + "".join([("\n\t\t " + k + '="' + v + '"') for k, v in d.items()])
            + ">"
        )

        # Author
        if "Author" in info_dictionary:
            s += (
                "\n\t\t\t<dc:creator>\n\t\t\t\t<rdf:Seq>\n\t\t\t\t\t<rdf:li>"
                + str(info_dictionary["Author"])
                + "</rdf:li>\n\t\t\t\t</rdf:Seq>\n\t\t\t</dc:creator>"
            )
        # Keywords
        if "Keywords" in info_dictionary:
            s += (
                "\n\t\t\t<dc:subject>\n\t\t\t\t<rdf:Bag>"
                + "".join(
                    [
                        ("\n\t\t\t\t\t<rdf:li>" + x.strip() + "</rdf:li>")
                        for x in str(info_dictionary["Keywords"]).split(" ")
                    ]
                )
                + "\n\t\t\t\t</rdf:Bag>\n\t\t\t</dc:subject>"
            )
        # Subject
        # fmt: off
        if "Subject" in info_dictionary:
            s += '\n\t\t\t<dc:description>\n\t\t\t\t<rdf:Alt>\n\t\t\t\t\t<rdf:li xml:lang="x-default">' + str(info_dictionary["Subject"]) + "</rdf:li>\n\t\t\t\t</rdf:Alt>\n\t\t\t</dc:description>"
        # fmt: on

        # Title
        # fmt: off
        if "Title" in info_dictionary:
            s += '\n\t\t\t<dc:title>\n\t\t\t\t<rdf:Alt>\n\t\t\t\t\t<rdf:li xml:lang="x-default">' + str(info_dictionary["Title"]) + "</rdf:li>\n\t\t\t\t</rdf:Alt>\n\t\t\t</dc:title>"
        # fmt: on

        # close
        s += "\n\t\t</rdf:Description>"

        # version
        if conformance_level is not None:
            s += '\n\t\t<rdf:Description rdf:about="" xmlns:pdfaid="http://www.aiim.org/pdfa/ns/id/">'
            s += (
                "\n\t\t\t<pdfaid:part>%d</pdfaid:part>"
                % conformance_level.get_standard()
            )
            s += (
                "\n\t\t\t<pdfaid:conformance>%s</pdfaid:conformance>"
                % conformance_level.get_conformance_level()
            )
            s += "\n\t\t</rdf:Description>"

        # close package
        s += "\n\t</rdf:RDF>"
        s += "\n</x:xmpmeta>"
        s += '\n<?xpacket end="w"?>'

        # build Stream object
        metadata_stream: Stream = Stream()
        metadata_stream[Name("Type")] = Name("Metadata")
        metadata_stream[Name("Subtype")] = Name("XML")
        metadata_stream[Name("Bytes")] = bytes(s, "latin1")
        metadata_stream[Name("Length")] = bDecimal(len(metadata_stream[Name("Bytes")]))

        # return
        return metadata_stream

    def transform(
        self,
        object_to_transform: Any,
        context: Optional[WriteTransformerState] = None,
    ):
        """
        This method writes an /Info Dictionary to a byte stream
        """

        # get Document
        document: Document = object_to_transform.get_root()
        assert document is not None
        assert isinstance(document, Document)

        # consolidate XMP /Metadata and /Info Dictionary
        new_info_dictionary: Dictionary = self._consolidate_xmp_and_info_dictionary(
            document
        )

        # update
        self._update_info_dictionary(new_info_dictionary)

        # determine whether XMP /Metadata is needed
        # fmt: off
        has_xmp_metadata: bool = "XRef" in document                                         \
                                   and "Trailer" in document["XRef"]                        \
                                   and "Root" in document["XRef"]["Trailer"]                \
                                   and "Metadata" in document["XRef"]["Trailer"]["Root"]
        needs_xmp_metadata = has_xmp_metadata or (context is not None and context.conformance_level is not None)
        # fmt: on

        if needs_xmp_metadata:

            # write XMP /Metadata
            xmp_metadata_stream: Stream = self._write_xmp_metadata_stream(
                new_info_dictionary, context.conformance_level
            )
            assert context is not None
            document["XRef"]["Trailer"]["Root"][Name("Metadata")] = self.get_reference(
                xmp_metadata_stream, context
            )
            xmp_metadata_stream.set_parent(document["XRef"]["Trailer"]["Root"])  # type: ignore [attr-defined]

            # delegate XMP /Metadata
            for h in self.get_root_transformer()._handlers:
                if isinstance(h, StreamTransformer) and h.can_be_transformed(
                    xmp_metadata_stream
                ):
                    h.transform(xmp_metadata_stream, context)
                    break

        # write /Info
        for k, v in new_info_dictionary.items():
            document["XRef"]["Trailer"][Name("Info")][k] = v

        # delegate /Info
        for h in self.get_root_transformer()._handlers:
            if isinstance(h, DictionaryTransformer) and h.can_be_transformed(
                document["XRef"]["Trailer"][Name("Info")]
            ):
                h.transform(document["XRef"]["Trailer"][Name("Info")], context)
                break

        # check reference
        assert document["XRef"]["Trailer"][Name("Info")].get_reference() is not None
        assert (
            document["XRef"]["Trailer"][Name("Info")].get_reference().byte_offset
            is not None
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of BaseTransformer is responsible for reading a Font object
"""
import io
import typing
from typing import Any, Optional, Union

from borb.io.read.transformer import ReadTransformerState, Transformer
from borb.io.read.types import AnyPDFType, Dictionary, Stream
from borb.pdf.canvas.event.event_listener import EventListener
from borb.pdf.canvas.font.composite_font.cid_font_type_0 import CIDType0Font
from borb.pdf.canvas.font.composite_font.cid_font_type_2 import CIDType2Font
from borb.pdf.canvas.font.composite_font.font_type_0 import Type0Font
from borb.pdf.canvas.font.font import Font
from borb.pdf.canvas.font.simple_font.font_type_1 import StandardType1Font, Type1Font
from borb.pdf.canvas.font.simple_font.font_type_3 import Type3Font
from borb.pdf.canvas.font.simple_font.true_type_font import TrueTypeFont


class FontDictionaryTransformer(Transformer):
    """
    This implementation of ReadBaseTransformer is responsible for reading a Font object
    """

    def __init__(self):
        super(FontDictionaryTransformer, self).__init__()
        self._accept_true_type_standard_14_fonts: bool = True

    def can_be_transformed(
        self, object: Union[io.BufferedIOBase, io.RawIOBase, io.BytesIO, AnyPDFType]
    ) -> bool:
        """
        This function returns True if the object to be transformed is a /Font Dictionary
        """
        return (
            isinstance(object, dict)
            and not isinstance(object, Stream)
            and "Type" in object
            and object["Type"] == "Font"
        )

    def transform(
        self,
        object_to_transform: Union[io.BufferedIOBase, io.RawIOBase, AnyPDFType],
        parent_object: Any,
        context: Optional[ReadTransformerState] = None,
        event_listeners: typing.List[EventListener] = [],
    ) -> Any:
        """
        This function reads a /Font Dictionary from a byte stream
        """

        # convert dictionary like structure
        assert isinstance(object_to_transform, Dictionary)
        subtype_name = object_to_transform["Subtype"]

        font_obj: Optional[Font] = None

        # TrueType Font
        if subtype_name == "TrueType":
            font_obj = TrueTypeFont()
            # Some libraries prefer to mark the standard 14 fonts as TrueType fonts
            # These libraries should be punished for their crimes against humanity.
            # But in the interest of pleasing the user, we explicitly catch these errors and guide
            # those wayward lambs back into the fold.
            if (
                self._accept_true_type_standard_14_fonts
                and "BaseFont" in object_to_transform
                and StandardType1Font.is_standard_14_font_name(
                    str(object_to_transform["BaseFont"])
                )
            ):
                font_obj = StandardType1Font(str(object_to_transform["BaseFont"]))

        # Type 0 Font
        elif subtype_name == "Type0":
            font_obj = Type0Font()

        # Type 1 Font
        elif subtype_name == "Type1":
            base_font: str = str(object_to_transform["BaseFont"])
            if StandardType1Font.is_standard_14_font_name(base_font):
                font_obj = StandardType1Font(base_font)
            else:
                font_obj = Type1Font()

        # Type 3 Font
        elif subtype_name == "Type3":
            font_obj = Type3Font()

        elif subtype_name == "CIDFontType0":
            font_obj = CIDType0Font()
        elif subtype_name == "CIDFontType2":
            font_obj = CIDType2Font()
        else:
            font_obj = StandardType1Font("Helvetica")

        # None
        if font_obj is None:
            return None

        # set parent
        assert font_obj is not None
        font_obj.set_parent(parent_object)  # type: ignore [union-attr]

        # convert key/value pair(s)
        assert isinstance(object_to_transform, Dictionary)
        for k, v in object_to_transform.items():
            if k == "Parent":
                continue
            v = self.get_root_transformer().transform(v, font_obj, context, [])
            if v is not None:
                font_obj[k] = v

        # return
        assert isinstance(font_obj, Font)
        return font_obj



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Append a rectangle to the current path as a complete
subpath, with lower-left corner (x, y) and dimensions width
and height in user space. The operation
x y width height re
is equivalent to
x y m
( x + width ) y l
( x + width ) ( y + height ) l
x ( y + height ) l
h
"""
import typing
from decimal import Decimal

from borb.io.read.types import AnyPDFType
from borb.io.read.types import Decimal as bDecimal
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class AppendRectangle(CanvasOperator):
    """
    Append a rectangle to the current path as a complete
    subpath, with lower-left corner (x, y) and dimensions width
    and height in user space. The operation
    x y width height re
    is equivalent to
    x y m
    ( x + width ) y l
    ( x + width ) ( y + height ) l
    x ( y + height ) l
    h
    """

    def __init__(self):
        super().__init__("s", 0)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the s operator
        """
        assert isinstance(operands[0], Decimal)
        assert isinstance(operands[1], Decimal)
        assert isinstance(operands[2], Decimal)
        assert isinstance(operands[3], Decimal)
        x: Decimal = operands[0]
        y: Decimal = operands[1]
        width: Decimal = operands[2]
        height: Decimal = operands[3]

        moveto_op: typing.Optional[
            CanvasOperator
        ] = canvas_stream_processor.get_operator("m")
        assert moveto_op is not None
        moveto_op.invoke(
            canvas_stream_processor, [bDecimal(x), bDecimal(y)], event_listeners
        )

        line_to_op: typing.Optional[
            CanvasOperator
        ] = canvas_stream_processor.get_operator("l")
        assert line_to_op is not None
        line_to_op.invoke(
            canvas_stream_processor, [bDecimal(x + width), bDecimal(y)], event_listeners
        )
        line_to_op.invoke(
            canvas_stream_processor,
            [bDecimal(x + width), bDecimal(y + height)],
            event_listeners,
        )
        line_to_op.invoke(
            canvas_stream_processor,
            [bDecimal(x), bDecimal(y + height)],
            event_listeners,
        )

        close_subpath_op: typing.Optional[
            CanvasOperator
        ] = canvas_stream_processor.get_operator("h")
        assert close_subpath_op is not None
        close_subpath_op.invoke(canvas_stream_processor, [], event_listeners)
